# =============================================================================
# Copyright (c) 2020 Tom Kralidis
#
# Author: Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

from io import BytesIO
import logging
from typing import BinaryIO

from owslib.ogcapi import Collections
from owslib.util import Authentication

LOGGER = logging.getLogger(__name__)


class Coverages(Collections):
    """Abstraction for OGC API - Coverages"""

    def __init__(self, url: str, json_: str = None, timeout: int = 30,
                 headers: dict = None, auth: Authentication = None):
        __doc__ = Collections.__doc__  # noqa
        super().__init__(url, json_, timeout, headers, auth)

    def coverages(self) -> list:
        """
        implements /collections filtered on coverages

        @returns: `list` of filtered collections object
        """

        coverages_ = []
        collections_ = super().collections()

        for c_ in collections_['collections']:
            for l_ in c_['links']:
                if 'coverage' in l_['rel']:
                    coverages_.append(c_['id'])
                    break

        return coverages_

    def coverage(self, collection_id: str, **kwargs: dict) -> BinaryIO:
        """
        implements /collection/{collectionId}/coverage/

        @type collection_id: string
        @param collection_id: id of collection
        @type properties: tuple | list
        @param properties: range subset
        @type subset: list or dict of tuples/lists
        @param subset:
            [(name, lower bound, upper bound)]
            [[name, lower bound, upper bound]]
            {name: (lower bound, upper bound)}
            {name: [lower bound, upper bound]}
        @type scale_size: list of tuples or dict
        @param scale_size: [(axis name, number)] | {axis name: number}
        @type scale_factor: int
        @param scale_factor: factor by which to scale the resulting coverage
        @type scale_axes: list of tuples or dict
        @param scale_axes: [(axis name, number)] | {axis name: number}
        @type datetime: tuple | list | str
        @param datetime:
            tuple or list of start/end datetimes, or as 'start/end' string
            start and end datetimes can be ".." for unbounded value

        @returns: coverage data
        """

        kwargs_ = {}

        if isinstance(kwargs.get('properties'), (tuple, list)):
            kwargs_['properties'] = ','.join([
                str(x) for x in kwargs['properties']
            ])

        for p in ['scale_axes', 'scale_size']:
            if p in kwargs:
                p2 = p.replace('_', '-')
                if isinstance(kwargs[p], (tuple, list)):
                    items = kwargs[p]
                elif isinstance(kwargs[p], dict):
                    items = [
                        (name, value)
                        for name, value
                        in kwargs[p].items()
                    ]
                else:
                    continue
                kwargs_[p2] = []
                kwargs_[p2] = ",".join(
                    f'{s[0]}({s[1]})'
                    for s in items
                )

        if 'scale_factor' in kwargs:
            scale_f = float(kwargs['scale_factor'])
            scale_i = int(scale_f)
            if scale_i == scale_f:
                kwargs_['scale-factor'] = scale_i
            else:
                kwargs_['scale-factor'] = scale_f

        if 'subset' in kwargs:
            subset_items = []
            subset_values = kwargs['subset']
            if isinstance(subset_values, (tuple, list)):
                subset_items = subset_values
            elif isinstance(subset_values, dict):
                subset_items = [
                    (name, *values)
                    for name, values
                    in subset_values.items()
                ]
            if subset_items:
                kwargs_['subset'] = ','.join([
                    f'{s[0]}({s[1]}:{s[2]})'
                    for s in subset_items
                ])

        if 'datetime' in kwargs:
            if isinstance(kwargs['datetime'], (tuple, list)):
                kwargs_['datetime'] = '/'.join(kwargs['datetime'][:2])
            else:
                kwargs_['datetime'] = str(kwargs['datetime'])

        path = f'collections/{collection_id}/coverage'

        return BytesIO(self._request(path=path, as_dict=False, kwargs=kwargs_))



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2018 Luca Cinquini
#
# Authors : Luca Cinquini <luca.cinquini@jpl.nasa.gov>
#           Carsten Ehbrecht <ehbrecht@dkrz.de>
#
# Contact email: ehbrecht@dkrz.de
# =============================================================================
"""
Abstract
--------

The wps module of the OWSlib package provides client-side functionality for executing invocations to a remote
Web Processing Server.


Disclaimer
----------

.. warning:: The owslib wps module should be considered in beta state: it has been tested versus only a handful of
    WPS services (deployed by the USGS, BADC and PML).
    More extensive testing is needed and feedback is appreciated.


Usage
-----

The module can be used to execute three types of requests versus a remote WPS endpoint:

# "GetCapabilities"

    * use the method wps.getcapabilities(xml=None)
    * the optional keyword argument "xml" may be used to avoid a real live request, and instead read the
      WPS capabilities document from a cached XML file

# "DescribeProcess"

    * use the method wps.describeprocess(identifier, xml=None)
    * identifier is the process identifier, retrieved from the list obtained from a previous
      "GetCapabilities" invocation
    * the optional keyword argument "xml" may be used to avoid a real live request, and instead read the
      WPS process description document from a cached XML file

# "Execute"

    * use the method wps.execute(identifier, inputs, output=None, request=None, response=None),
      which submits the job to the remote WPS server and returns a WPSExecution object that can be used to periodically
      check the job status until completion (or error)
    * the optional keyword argument "request" may be used to avoid re-building the request XML from input arguments,
      and instead submit a request from a pre-made XML file
    * alternatively, an "Execute" request can be built from input arguments by supplying the "identifier", "inputs"
      and "output" arguments to the execute() method.

        * "identifier" is the mandatory process identifier
        * "inputs" is a dictionary of (key,value) pairs where:

            * key is a named input parameter
            * value is either a string, or any Python object that supports a getXml() method
              In particular, a few classes are included in the package to support a FeatuteCollection input:

                  * "WFSFeatureCollection" can be used in conjunction with "WFSQuery" to define a FEATURE_COLLECTION
                    retrieved from a live WFS server.
                  * "GMLMultiPolygonFeatureCollection" can be used to define one or more polygons
                    of (latitude, longitude) points.
          * "output" is an optional output identifier to be included in the ResponseForm section of the request.

    * the optional keyword argument "response" mey be used to avoid submitting a real live request, and instead
      reading the WPS execution response document from a cached XML file (for debugging or testing purposes)
    * the convenience module function monitorExecution() can be used to periodically check the status of a remote
      running job, and eventually download the output either to a named file, or to a file specified by the server.


Examples
--------

The files examples/wps-usgs-script.py, examples/wps-pml-script-1.py and examples/wps-pml-script-2.py contain
real-world usage examples that submits a "GetCapabilities", "DescribeProcess" and "Execute" requests to
the live USGS and PML servers. To run:

* cd examples
* python3 wps-usgs-script.py
* python3 wps-pml-script-1.py
* python3 wps-pml-script-2.py

The file wps-client.py contains a command-line client that can be used to submit a "GetCapabilities",
"DescribeProcess" or "Execute" request to an arbitrary WPS server. For example, you can run it as follows:

* cd examples
* To prints out usage and example invocations: wps-client -help
* To execute a (fake) WPS invocation::

      $ wps-client.py -v -u http://cida.usgs.gov/climate/gdp/process/WebProcessingService -r GetCapabilities -x ../tests/USGSCapabilities.xml  # noqa

The directory tests/ includes several doctest-style files wps_*.txt that show how to interactively submit a
"GetCapabilities", "DescribeProcess" or "Execute" request, without making a live request but rather parsing the
response of cached XML response documents. To run:

* cd tests
* python3 -m doctest wps_*.txt
    ``(or python3 -m doctest -v wps_*.txt for verbose output)``

Also, the directory tests/ contains several examples of well-formed "Execute" requests:

* The files wps_USGSExecuteRequest*.xml contain requests that can be submitted to the live USGS WPS service.
* The files PMLExecuteRequest*.xml contain requests that can be submitted to the live PML WPS service.
"""

import logging

from owslib.etree import etree
from owslib.ows import DEFAULT_OWS_NAMESPACE, XLINK_NAMESPACE
from owslib.ows import ServiceIdentification, ServiceProvider, OperationsMetadata, BoundingBox
from time import sleep
from owslib.util import (testXMLValue, testXMLAttribute, build_get_url, clean_ows_url, dump, getTypedValue,
                         getNamespace, element_to_string, nspath, openURL, nspath_eval, Authentication)
from xml.dom.minidom import parseString
from owslib.namespaces import Namespaces
from urllib.parse import urlparse
import warnings
import os

# namespace definition
n = Namespaces()

# These static namespaces are DEPRECIATED.  Please don't use them.
# No great way of printing a message since there are at the file level
WPS_DEFAULT_NAMESPACE = n.get_namespace("wps")
WFS_NAMESPACE = n.get_namespace("wfs")
OGC_NAMESPACE = n.get_namespace("ogc")
GML_NAMESPACE = n.get_namespace("gml")
DRAW_NAMESPACE = n.get_namespace("draw")

GML_SCHEMA_LOCATION = "http://schemas.opengis.net/gml/3.1.1/base/feature.xsd"
DRAW_SCHEMA_LOCATION = 'http://cida.usgs.gov/climate/derivative/xsd/draw.xsd'
WFS_SCHEMA_LOCATION = 'http://schemas.opengis.net/wfs/1.1.0/wfs.xsd'
WPS_DEFAULT_SCHEMA_LOCATION = 'http://schemas.opengis.net/wps/1.0.0/wpsExecute_request.xsd'
WPS_DEFAULT_VERSION = '1.0.0'

# WPS execution modes
AUTO = 'auto'
SYNC = 'sync'
ASYNC = 'async'

LOGGER = logging.getLogger(__name__)


def get_namespaces():
    ns = n.get_namespaces(["ogc", "wfs", "wps", "gml", "xsi", "xlink"])
    ns[None] = n.get_namespace("wps")
    ns["ows"] = DEFAULT_OWS_NAMESPACE
    return ns


namespaces = get_namespaces()


def is_reference(val):
    """
    Checks if the provided value is a reference (URL).
    """
    try:
        parsed = urlparse(val)
        is_ref = bool(parsed.scheme)
    except Exception:
        is_ref = False
    return is_ref


def is_literaldata(val):
    """
    Checks if the provided value is a string (includes unicode).
    """
    return isinstance(val, str)


def is_boundingboxdata(val):
    """
    Checks if the provided value is an implementation of ``BoundingBoxDataInput``.
    """
    return isinstance(val, BoundingBoxDataInput)


def is_complexdata(val):
    """
    Checks if the provided value is an implementation of ``IComplexDataInput``.
    """
    return isinstance(val, IComplexDataInput)


def _fix_auth(auth, username=None, password=None, verify=None, cert=None):
    """Updates auth from deprecated parameters username, password, verify and cert."""
    if any(p is not None for p in (username, password, verify, cert)):
        message = 'The use of "username", "password", "verify", and "cert" is deprecated. ' + \
                  'Please use the "auth" keyword during class instantiation. ' + \
                  'These keywords will be removed in a future release.'
        warnings.warn(message, DeprecationWarning)
    if username is not None:
        auth.username = username
    if password is not None:
        auth.password = password
    if verify is not None:
        auth.verify = verify
    if cert is not None:
        auth.cert = cert
    return auth


class IComplexDataInput(object):

    """
    Abstract interface representing complex input object for a WPS request.
    """

    def getXml(self):
        """
        Method that returns the object data as an XML snippet,
        to be inserted into the WPS request document sent to the server.
        """
        raise NotImplementedError


class WebProcessingService(object):

    """
    Class that contains client-side functionality for invoking an OGC Web Processing Service (WPS).

    Implements IWebProcessingService.
    """

    def __init__(self, url, version=WPS_DEFAULT_VERSION, username=None, password=None, skip_caps=False,
                 headers=None, verify=None, cert=None, timeout=None, auth=None, language=None):
        """
        Initialization method resets the object status.
        By default it will execute a GetCapabilities invocation to the remote service,
        which can be skipped by using skip_caps=True.

        Parameters username, password, verify and cert are deprecated. Please use auth parameter.
        """
        self.auth = auth or Authentication()
        _fix_auth(self.auth, username, password, verify, cert)

        # fields passed in from object initializer
        self.url = clean_ows_url(url)
        self.version = version
        self.headers = headers
        self.timeout = timeout
        self.language = language

        # fields populated by method invocations
        self._capabilities = None
        self.identification = None
        self.provider = None
        self.operations = []
        self.processes = []
        self.languages = None

        if not skip_caps:
            self.getcapabilities()

    def getcapabilities(self, xml=None):
        """
        Method that requests a capabilities document from the remote WPS server and populates this object's metadata.
        keyword argument xml: local XML GetCapabilities document, prevents actual HTTP invocation.
        """

        # read capabilities document
        reader = WPSCapabilitiesReader(
            version=self.version,
            auth=self.auth,
            language=self.language,
            timeout=self.timeout,
        )
        if xml:
            # read from stored XML file
            self._capabilities = reader.readFromString(xml)
        else:
            self._capabilities = reader.readFromUrl(
                self.url, headers=self.headers)

        LOGGER.debug(element_to_string(self._capabilities))

        # populate the capabilities metadata obects from the XML tree
        self._parseCapabilitiesMetadata(self._capabilities)

    def describeprocess(self, identifier, xml=None):
        """
        Requests a process document from a WPS service and populates the process metadata.
        Returns the process object or a list of process objects.

        :param str identifier: The process id. If `all`, return a list of all processes available.
        """

        # read capabilities document
        reader = WPSDescribeProcessReader(
            version=self.version,
            auth=self.auth,
            language=self.language,
            timeout=self.timeout,
        )
        if xml:
            # read from stored XML file
            rootElement = reader.readFromString(xml)
        else:
            # read from server
            rootElement = reader.readFromUrl(
                self.url, identifier, headers=self.headers)

        LOGGER.info(element_to_string(rootElement))

        # build metadata objects
        processes = self._parseProcessMetadata(rootElement)

        if identifier == 'all':
            return processes
        # return process with given identifier
        for process in processes:
            if process.identifier == identifier:
                return process
        raise ValueError('process with identifier {} not found'.format(identifier))

    def execute(self, identifier, inputs, output=None, mode=ASYNC, lineage=False, request=None, response=None):
        """
        Submits a WPS process execution request.
        Returns a WPSExecution object, which can be used to monitor the status of the job, and ultimately
        retrieve the result.

        :param str identifier: the requested process identifier
        :param inputs: list of process inputs as (input_identifier, value) tuples (where value is either a string
                for LiteralData, or an object for ComplexData).
        :param output: optional list of process outputs as tuples (output_identifier, as_ref, mime_type).
                `as_ref` can be True (as reference),
                False (embedded in response) or None (use service default).
                `mime_type` should be text or None (use service default)
        :param mode: execution mode: SYNC, ASYNC or AUTO. Default: ASYNC
        :param lineage: if lineage is "true", the Execute operation response shall include the DataInputs and
                 OutputDefinitions elements.
        :param request: optional pre-built XML request document, prevents building of request from other arguments
        :param response: optional pre-built XML response document, prevents submission of request to live WPS server
        """

        # instantiate a WPSExecution object
        LOGGER.info('Executing WPS request...')
        execution = WPSExecution(
            version=self.version,
            url=self.url,
            headers=self.headers,
            timeout=self.timeout,
            auth=self.auth,
            language=self.language,
        )

        # build XML request from parameters
        if request is None:
            requestElement = execution.buildRequest(identifier, inputs, output, mode=mode, lineage=lineage)
            request = etree.tostring(requestElement)
            execution.request = request
        LOGGER.debug(request)

        # submit the request to the live server
        if response is None:
            response = execution.submitRequest(request)
        else:
            response = etree.fromstring(response)

        LOGGER.debug(etree.tostring(response))

        # parse response
        execution.parseResponse(response)

        return execution

    def getOperationByName(self, name):
        """Return a named content item."""
        for item in self.operations:
            if item.name == name:
                return item
        raise KeyError("No operation named %s" % name)

    def _parseProcessMetadata(self, rootElement):
        """Return a list of Process objects parsed from a <ProcessDescriptions> XML element."""

        processDescriptionElements = rootElement.findall('ProcessDescription')
        processes = []
        for processDescriptionElement in processDescriptionElements:
            process = Process(processDescriptionElement)

            # override existing processes in object metadata, if existing already
            found = False
            for n, p in enumerate(self.processes):
                if p.identifier == process.identifier:
                    self.processes[n] = process
                    found = True
            # otherwise add it
            if not found:
                self.processes.append(process)

            processes.append(process)

        return processes

    def _parseCapabilitiesMetadata(self, root):
        """Set up capabilities metadata objects."""

        # reset metdata
        self.operations = []
        self.processes = []

        # use the WPS namespace defined in the document root
        # TODO: wpsns not used
        # wpsns = getNamespace(root)

        self.updateSequence = root.attrib.get('updateSequence')

        # loop over children WITHOUT requiring a specific namespace
        for element in root:

            # this element's namespace
            ns = getNamespace(element)

            # <ows:ServiceIdentification> metadata
            if element.tag.endswith('ServiceIdentification'):
                self.identification = ServiceIdentification(
                    element, namespace=ns)
                LOGGER.debug(self.identification)

            # <ows:ServiceProvider> metadata
            elif element.tag.endswith('ServiceProvider'):
                self.provider = ServiceProvider(element, namespace=ns)
                LOGGER.debug(self.provider)

            # <ns0:OperationsMetadata xmlns:ns0="http://www.opengeospatial.net/ows">
            #   <ns0:Operation name="GetCapabilities">
            #     <ns0:DCP>
            #       <ns0:HTTP>
            #         <ns0:Get xlink:href="http://ceda-wps2.badc.rl.ac.uk/wps?" xmlns:xlink="http://www.w3.org/1999/xlink" />  # noqa
            #       </ns0:HTTP>
            #    </ns0:DCP>
            #  </ns0:Operation>
            #  ........
            # </ns0:OperationsMetadata>
            elif element.tag.endswith('OperationsMetadata'):
                for child in element.findall(nspath('Operation', ns=ns)):
                    self.operations.append(
                        OperationsMetadata(child, namespace=ns))
                    LOGGER.debug(self.operations[-1])

            # <wps:ProcessOfferings>
            #   <wps:Process ns0:processVersion="1.0.0">
            #     <ows:Identifier xmlns:ows="http://www.opengis.net/ows/1.1">gov.usgs.cida.gdp.wps.algorithm.filemanagement.ReceiveFiles</ows:Identifier>  # noqa
            #     <ows:Title xmlns:ows="http://www.opengis.net/ows/1.1">gov.usgs.cida.gdp.wps.algorithm.filemanagement.ReceiveFiles</ows:Title>            # noqa
            #   </wps:Process>
            #   ......
            # </wps:ProcessOfferings>
            elif element.tag.endswith('ProcessOfferings'):
                for child in element.findall(nspath('Process', ns=ns)):
                    p = Process(child)
                    self.processes.append(p)
                    LOGGER.debug(self.processes[-1])

            # <wps:Languages>
            #   <wps:Default>
            #     <ows:Language>en-US</ows:Language>
            #   </wps:Default>
            #   <wps:Supported>
            #     <ows:Language>en-US</ows:Language>
            #     <ows:Language>fr-CA</ows:Language>
            #     ......
            #   </wps:Supported>
            # </wps:Languages>
            elif element.tag.endswith('Languages'):
                self.languages = Languages(element)
                LOGGER.debug(self.languages)


class WPSReader(object):
    """
    Superclass for reading a WPS document into a lxml.etree infoset.
    """

    def __init__(self, version=WPS_DEFAULT_VERSION, timeout=30, auth=None, language=None):
        self.version = version
        self.timeout = timeout
        self.auth = auth or Authentication()
        self.language = language

    def _readFromUrl(self, url, data, timeout, method='Get', username=None, password=None,
                     headers=None, verify=True, cert=None):
        """
        Method to get and parse a WPS document, returning an elementtree instance.
        :param str url: WPS service base url.
        :param {} data: GET: dictionary of HTTP (key, value) parameter pairs, POST: XML document to post
        """
        _fix_auth(self.auth, username, password, verify, cert)
        if method == 'Get':
            # full HTTP request url
            if self.language:
                data["language"] = self.language
            request_url = build_get_url(url, data, overwrite=True)
            LOGGER.debug(request_url)

            # split URL into base url and query string to use utility function
            spliturl = request_url.split('?')
            u = openURL(spliturl[0], spliturl[
                        1], method='Get', username=self.auth.username, password=self.auth.password,
                        headers=headers, verify=self.auth.verify, cert=self.auth.cert, timeout=self.timeout)
            return etree.fromstring(u.read())

        elif method == 'Post':
            u = openURL(url, data, method='Post',
                        username=self.auth.username, password=self.auth.password,
                        headers=headers, verify=self.auth.verify, cert=self.auth.cert, timeout=timeout)
            return etree.fromstring(u.read())

        else:
            raise Exception("Unrecognized HTTP method: %s" % method)

    def readFromString(self, string):
        """
        Method to read a WPS GetCapabilities document from an XML string.
        """

        if not isinstance(string, str) and not isinstance(string, bytes):
            raise ValueError(
                "Input must be of type string, not %s" % type(string))
        return etree.fromstring(string)


class WPSCapabilitiesReader(WPSReader):

    """
    Utility class that reads and parses a WPS GetCapabilities document into a lxml.etree infoset.
    """

    def __init__(self, version=WPS_DEFAULT_VERSION, timeout=None, auth=None, language=None):
        # superclass initializer
        super(WPSCapabilitiesReader, self).__init__(
            version=version, timeout=timeout, auth=auth, language=language)

    def readFromUrl(self, url, username=None, password=None,
                    headers=None, verify=None, cert=None):
        """
        Method to get and parse a WPS capabilities document, returning an elementtree instance.

        :param str url: WPS service base url, to which is appended the HTTP parameters: service, version, and request.
        """
        return self._readFromUrl(url,
                                 {'service': 'WPS', 'request':
                                     'GetCapabilities', 'version': self.version},
                                 self.timeout,
                                 username=username, password=password,
                                 headers=headers, verify=verify, cert=cert)


class WPSDescribeProcessReader(WPSReader):

    """
    Class that reads and parses a WPS DescribeProcess document into a etree infoset
    """

    def __init__(self, version=WPS_DEFAULT_VERSION, timeout=None, auth=None, language=None):
        # superclass initializer
        super(WPSDescribeProcessReader, self).__init__(
            version=version, timeout=timeout, auth=auth, language=language)

    def readFromUrl(self, url, identifier, username=None, password=None,
                    headers=None, verify=None, cert=None):
        """
        Reads a WPS DescribeProcess document from a remote service and returns the XML etree object

        :param str url: WPS service base url, to which is appended the HTTP parameters: 'service', 'version',
            'request', and 'identifier'.
        """
        return self._readFromUrl(url,
                                 {'service': 'WPS', 'request': 'DescribeProcess',
                                     'version': self.version, 'identifier': identifier},
                                 self.timeout,
                                 username=username, password=password,
                                 headers=headers, verify=verify, cert=cert)


class WPSExecuteReader(WPSReader):

    """
    Class that reads and parses a WPS Execute response document into a etree infoset
    """

    def __init__(self, timeout=None, auth=None, language=None):
        # superclass initializer
        super(WPSExecuteReader, self).__init__(timeout=timeout, auth=auth, language=language)

    def readFromUrl(self, url, data={}, method='Get', username=None, password=None,
                    headers=None, verify=None, cert=None):
        """
        Reads a WPS status document from a remote service and returns the XML etree object.
        :param str url: the URL to submit the GET/POST request to.
        """

        return self._readFromUrl(url, data, self.timeout, method, username=username, password=password,
                                 headers=headers, verify=verify, cert=cert)


class WPSExecution(object):

    """
    Class that represents a single WPS process executed on a remote WPS service.
    """

    def __init__(self, version=WPS_DEFAULT_VERSION, url=None, username=None, password=None,
                 headers=None, verify=None, cert=None, timeout=None, auth=None, language=None):

        # initialize fields
        self.url = url
        self.version = version
        self.headers = headers
        self.auth = auth or Authentication()
        _fix_auth(self.auth, username, password, verify, cert)
        self.timeout = timeout
        self.language = language

        # request document
        self.request = None

        # last response document
        self.response = None

        # status fields retrieved from the response documents
        self.process = None
        self.serviceInstance = None
        self.status = None
        self.percentCompleted = 0
        self.statusMessage = None
        self.errors = []
        self.statusLocation = None
        self.dataInputs = []
        self.processOutputs = []
        self.creationTime = None

    def buildRequest(self, identifier, inputs=[], output=None, mode=ASYNC, lineage=False):
        """
        Method to build a WPS process request.
        :param str identifier: the requested process identifier
        :param inputs: array of input arguments for the process.
            - LiteralData inputs are expressed as simple (key,value) tuples where key is the input identifier,
              value is the value
            - ComplexData inputs are expressed as (key, object) tuples, where key is the input identifier,
              and the object must contain a 'getXml()' method that returns an XML infoset to be included in
              the WPS request
        :param output: array of outputs which should be returned:
                expressed as tuples (key, as_ref, mime_mype) where key is the output identifier and as_ref is True
                if output should be returned as reference.
                as_ref and mimeType may be null for using server's default value
        :param mode: execution mode: SYNC, ASYNC or AUTO.
        :param lineage: if lineage is "true", the Execute operation response shall include the DataInputs and
                 OutputDefinitions elements.
        """
        # TODO: auto mode needs to implemented for WPS 2.0.0
        if mode is SYNC:
            _async = False
        elif mode is AUTO:
            LOGGER.warn("Auto mode not available in WPS 1.0.0. Using async mode.")
            _async = True
        else:
            _async = True

        # <wps:Execute xmlns:wps="http://www.opengis.net/wps/1.0.0"
        #             xmlns:ows="http://www.opengis.net/ows/1.1"
        #             xmlns:xlink="http://www.w3.org/1999/xlink"
        #             xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
        #             service="WPS"
        #             version="1.0.0"
        # xsi:schemaLocation="http://www.opengis.net/wps/1.0.0
        # http://schemas.opengis.net/wps/1.0.0/wpsExecute_request.xsd">
        root = etree.Element(nspath_eval('wps:Execute', namespaces))
        root.set('service', 'WPS')
        root.set('version', WPS_DEFAULT_VERSION)
        if self.language:
            root.set('language', self.language)
        root.set(nspath_eval('xsi:schemaLocation', namespaces), '%s %s' %
                 (namespaces['wps'], WPS_DEFAULT_SCHEMA_LOCATION))

        # <ows:Identifier>gov.usgs.cida.gdp.wps.algorithm.FeatureWeightedGridStatisticsAlgorithm</ows:Identifier>
        identifierElement = etree.SubElement(
            root, nspath_eval('ows:Identifier', namespaces))
        identifierElement.text = identifier

        # <wps:DataInputs>
        dataInputsElement = etree.SubElement(
            root, nspath_eval('wps:DataInputs', namespaces))

        for (key, val) in inputs:

            inputElement = etree.SubElement(
                dataInputsElement, nspath_eval('wps:Input', namespaces))
            identifierElement = etree.SubElement(
                inputElement, nspath_eval('ows:Identifier', namespaces))
            identifierElement.text = key

            # Literal data
            # <wps:Input>
            #   <ows:Identifier>DATASET_URI</ows:Identifier>
            #   <wps:Data>
            #     <wps:LiteralData>dods://igsarm-cida-thredds1.er.usgs.gov:8080/thredds/dodsC/dcp/conus_grid.w_meta.ncml</wps:LiteralData>
            #   </wps:Data>
            # </wps:Input>
            if is_literaldata(val):
                LOGGER.debug("literaldata %s", key)
                dataElement = etree.SubElement(
                    inputElement, nspath_eval('wps:Data', namespaces))
                literalDataElement = etree.SubElement(
                    dataElement, nspath_eval('wps:LiteralData', namespaces))
                literalDataElement.text = val

            # Complex data
            # <wps:Input>
            #   <ows:Identifier>FEATURE_COLLECTION</ows:Identifier>
            #   <wps:Reference xlink:href="http://igsarm-cida-gdp2.er.usgs.gov:8082/geoserver/wfs">
            #      <wps:Body>
            #        <wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs" xmlns:ogc="http://www.opengis.net/ogc" xmlns:gml="http://www.opengis.net/gml" service="WFS" version="1.1.0" outputFormat="text/xml; subtype=gml/3.1.1" xsi:schemaLocation="http://www.opengis.net/wfs ../wfs/1.1.0/WFS.xsd">  # noqa
            #            <wfs:Query typeName="sample:CONUS_States">
            #                <wfs:PropertyName>the_geom</wfs:PropertyName>
            #                <wfs:PropertyName>STATE</wfs:PropertyName>
            #                <ogc:Filter>
            #                    <ogc:GmlObjectId gml:id="CONUS_States.508"/>
            #                </ogc:Filter>
            #            </wfs:Query>
            #        </wfs:GetFeature>
            #      </wps:Body>
            #   </wps:Reference>
            # </wps:Input>
            elif is_complexdata(val):
                LOGGER.debug("complexdata %s", key)
                inputElement.append(val.getXml())
            elif is_boundingboxdata(val):
                inputElement.append(val.get_xml())
            else:
                raise Exception(
                    'input type of "%s" parameter is unknown' % key)

        # <wps:ResponseForm>
        #   <wps:ResponseDocument storeExecuteResponse="true" status="true" lineage="false">
        #     <wps:Output asReference="true" mimeType="application/json">
        #       <ows:Identifier>OUTPUT</ows:Identifier>
        #     </wps:Output>
        #   </wps:ResponseDocument>
        # </wps:ResponseForm>
        if output is not None:
            responseFormElement = etree.SubElement(
                root, nspath_eval('wps:ResponseForm', namespaces))
            responseDocumentElement = etree.SubElement(
                responseFormElement, nspath_eval(
                    'wps:ResponseDocument', namespaces),
                attrib={'storeExecuteResponse': str(_async).lower(),
                        'status': str(_async).lower(),
                        'lineage': str(lineage).lower()})
            # keeping backward compability of output parameter
            if isinstance(output, str):
                self._add_output(responseDocumentElement, output)
            elif isinstance(output, list):
                for ouputTuple in output:
                    # tuple (identifier, as_reference) for backward compatibility
                    if len(ouputTuple) == 2:
                        (identifier, as_reference) = ouputTuple
                        mime_type = None
                    else:
                        (identifier, as_reference, mime_type) = ouputTuple
                    self._add_output(
                        responseDocumentElement, identifier, asReference=as_reference, mimeType=mime_type)
            else:
                raise Exception(
                    'output parameter is neither string nor list. output=%s' % output)
        return root

    def _add_output(self, element, identifier, asReference=None, mimeType=None):
        output_element = etree.SubElement(
            element, nspath_eval('wps:Output', namespaces))
        if isinstance(mimeType, str):
            output_element.attrib['mimeType'] = mimeType
        if isinstance(asReference, bool):
            output_element.attrib['asReference'] = str(asReference).lower()
        # outputIdentifierElement
        etree.SubElement(
            output_element, nspath_eval('ows:Identifier', namespaces)).text = identifier

    # wait for 60 seconds by default
    def checkStatus(self, url=None, response=None, sleepSecs=60):
        """
        Method to check the status of a job execution.
        In the process, this method will update the object 'response' attribute.

        :param str url: optional 'statusLocation' URL retrieved from a previous WPS Execute response document.
             If not provided, the current 'statusLocation' URL will be used.
        :param int sleepSecs: number of seconds to sleep before returning control to the caller.
        """

        reader = WPSExecuteReader(
            auth=self.auth,
            language=self.language,
            timeout=self.timeout
        )
        if response is None:
            # override status location
            if url is not None:
                self.statusLocation = url
            LOGGER.info('\nChecking execution status... (location=%s)' %
                        self.statusLocation)
            try:
                response = reader.readFromUrl(
                    self.statusLocation, headers=self.headers)
            except Exception:
                LOGGER.error("Could not read status document.")
        else:
            response = reader.readFromString(response)

        # store latest response
        try:
            xml = etree.tostring(response)
        except Exception:
            LOGGER.error("Could not parse XML response.")
        else:
            self.response = xml
            LOGGER.debug(self.response)

            self.parseResponse(response)

            # sleep given number of seconds
            if self.isComplete() is False:
                LOGGER.info('Sleeping %d seconds...' % sleepSecs)
                sleep(sleepSecs)

    def getStatus(self):
        return self.status

    def isComplete(self):
        if (self.status == 'ProcessSucceeded' or self.status == 'ProcessFailed' or self.status == 'Exception'):
            return True
        elif (self.status == 'ProcessStarted'):
            return False
        elif (self.status == 'ProcessAccepted' or self.status == 'ProcessPaused'):
            return False
        else:
            raise Exception(
                'Unknown process execution status: %s' % self.status)

    def isSucceded(self):
        """
        Calls self.isSucceeded - deprecated because of spelling mistake,
        retained for backwards-compatibility.
        """
        return self.isSucceeded()

    def isSucceeded(self):
        """
        Returns a boolean regarding the success status of the process.
        """
        if self.status == 'ProcessSucceeded':
            return True
        else:
            return False

    def isNotComplete(self):
        return not self.isComplete()

    def getOutput(self, filepath=None, identifier=None):
        """
        Method to write the outputs of a WPS process to a file:
        either retrieves the referenced files from the server, or writes out the content of response embedded output.

        :param filepath: optional path to the output file, otherwise a file will be created in the local directory with
                  the name assigned by the server, or default name 'wps.out' for embedded output.
        :param: identifier: optional identifier of the output that should be written.
                  For backward compatibility it will default to the first output.
        """

        if self.isSucceeded():
            content = b''
            output = None
            if self.processOutputs:
                if identifier:
                    # filter outputs by identifier
                    outputs = [o for o in self.processOutputs if o.identifier == identifier]
                    if outputs:
                        output = outputs[0]
                else:
                    # take the first found output
                    output = self.processOutputs[0]
            if output:
                # ExecuteResponse contains reference to server-side output
                if output.reference:
                    content = output.retrieveData(
                        self.auth.username, self.auth.password,
                        headers=self.headers, verify=self.auth.verify, cert=self.auth.cert)
                    if filepath is None:
                        filepath = output.fileName
                # ExecuteResponse contain embedded output
                elif len(output.data) > 0:
                    if filepath is None:
                        filepath = 'wps.out'
                    for data in output.data:
                        content = content + data.encode()
            # write out content
            if content != b'':
                out = open(filepath, 'wb')
                out.write(content)
                out.close()
                LOGGER.info(f'Output written to file: {filepath}')
        else:
            raise Exception(
                f"Execution not successfully completed: status={self.status}")

    def submitRequest(self, request):
        """
        Submits a WPS Execute document to a remote service, returns the XML response document from the server.
        This method will save the request document and the first returned response document.

        :param request: the XML request document to be submitted as POST to the server.
        """

        self.request = request
        reader = WPSExecuteReader(
            auth=self.auth,
            language=self.language,
            timeout=self.timeout,
        )
        response = reader.readFromUrl(
            self.url, request, method='Post', headers=self.headers)
        self.response = response
        return response

        '''
        if response is None:
            # override status location
            if url is not None:
                self.statusLocation = url

        else:
            response = reader.readFromString(response)


        '''

    def parseResponse(self, response):
        """
        Method to parse a WPS response document
        """

        rootTag = response.tag.split('}')[1]
        # <ns0:ExecuteResponse>
        if rootTag == 'ExecuteResponse':
            self._parseExecuteResponse(response)

        # <ows:ExceptionReport>
        elif rootTag == 'ExceptionReport':
            self._parseExceptionReport(response)

        else:
            LOGGER.debug('Unknown Response')

        # log status, errors
        LOGGER.info('Execution status=%s' % self.status)
        LOGGER.info('Percent completed=%s' % self.percentCompleted)
        LOGGER.info('Status message=%s' % self.statusMessage)
        for error in self.errors:
            dump(error)

    def _parseExceptionReport(self, root):
        """
        Method to parse a WPS ExceptionReport document and populate this object's metadata.
        """
        # set exception status
        self.status = "Exception"

        for exceptionEl in root.findall(nspath('Exception', ns=namespaces['ows'])):
            self.errors.append(WPSException(exceptionEl))

    def _parseExecuteResponse(self, root):
        """
        Method to parse a WPS ExecuteResponse response document and populate this object's metadata.
        """

        # retrieve WPS namespace directly from root element
        wpsns = getNamespace(root)

        self.serviceInstance = root.get('serviceInstance')
        if self.statusLocation is None:
            self.statusLocation = root.get('statusLocation')

        # <ns0:Status creationTime="2011-11-09T14:19:50Z">
        #  <ns0:ProcessSucceeded>PyWPS Process v.net.path successfully calculated</ns0:ProcessSucceeded>
        # </ns0:Status>
        # OR
        # <ns0:Status creationTime="2011-11-07T08:26:44.359-06:00">
        #  <ns0:ProcessFailed>
        #   <ows:ExceptionReport xmlns:ows="http://www.opengis.net/ows/1.1">
        #    <ows:Exception>
        #     <ows:ExceptionText>Attribute null not found in feature collection</ows:ExceptionText>
        #    </ows:Exception>
        #   </ows:ExceptionReport>
        #  </ns0:ProcessFailed>
        # </ns0:Status>
        statusEl = root.find(nspath('Status/*', ns=wpsns))
        self.status = statusEl.tag.split('}')[1]
        # creationTime attribute
        element = root.find(nspath('Status', ns=wpsns))
        self.creationTime = testXMLAttribute(element, 'creationTime')
        # get progress info
        if self.status == 'ProcessSucceeded':
            self.percentCompleted = 100
        else:
            try:
                percentCompleted = int(statusEl.get('percentCompleted'))
                self.percentCompleted = percentCompleted
            except Exception:
                pass
        # get status message
        self.statusMessage = statusEl.text
        # exceptions ?
        for element in statusEl:
            if element.tag.endswith('ExceptionReport'):
                self._parseExceptionReport(element)

        self.process = Process(
            root.find(nspath('Process', ns=wpsns)))

        # <wps:DataInputs xmlns:wps="http://www.opengis.net/wps/1.0.0"
        # xmlns:ows="http://www.opengis.net/ows/1.1"
        # xmlns:xlink="http://www.w3.org/1999/xlink">
        if len(self.dataInputs) > 0:
            LOGGER.debug('clean data inputs')
            self.dataInputs[:] = []
        for inputElement in root.findall(nspath('DataInputs/Input', ns=wpsns)):
            self.dataInputs.append(Output(inputElement))
            LOGGER.debug(self.dataInputs[-1])

        # <ns:ProcessOutputs>
        # xmlns:ns="http://www.opengis.net/wps/1.0.0"
        if len(self.processOutputs) > 0:
            LOGGER.debug('clean process outputs')
            self.processOutputs[:] = []
        for outputElement in root.findall(nspath('ProcessOutputs/Output', ns=wpsns)):
            self.processOutputs.append(Output(outputElement))
            LOGGER.debug(self.processOutputs[-1])


class ComplexData(object):

    """
    Class that represents a ComplexData element in a WPS document
    """

    def __init__(self, mimeType=None, encoding=None, schema=None, maximumMegaBytes=None):
        self.mimeType = mimeType
        self.encoding = encoding
        self.schema = schema
        self.maximumMegabytes = maximumMegaBytes  # defined only if provided by ProcessDescription ComplexData


class InputOutput(object):

    """
    Superclass of a WPS input or output data object.
    """

    def __init__(self, element):

        self.abstract = None
        self.metadata = []

        # loop over sub-elements without requiring a specific namespace
        for child in element:

            # <ows:Identifier xmlns:ows="http://www.opengis.net/ows/1.1">SUMMARIZE_TIMESTEP</ows:Identifier>
            if child.tag.endswith('Identifier'):
                self.identifier = testXMLValue(child)

            # <ows:Title xmlns:ows="http://www.opengis.net/ows/1.1">Summarize Timestep</ows:Title>
            elif child.tag.endswith('Title'):
                self.title = testXMLValue(child)

            # <ows:Abstract xmlns:ows="http://www.opengis.net/ows/1.1">
            #   If selected, processing output will include columns with summarized statistics for all
            #   feature attribute values for each timestep
            # </ows:Abstract>
            elif child.tag.endswith('Abstract'):
                self.abstract = testXMLValue(child)

            # <ows:Metadata xlink:title="Documentation" xlink:href="http://emu.readthedocs.org/en/latest/"/>
            elif child.tag.endswith('Metadata'):
                self.metadata.append(Metadata(child))

        self.allowedValues = []
        self.supportedValues = []
        self.defaultValue = None
        self.dataType = None
        self.anyValue = False

    def _parseData(self, element):
        """
        Method to parse a "Data" element
        """

        # <ns0:Data>
        #        <ns0:ComplexData mimeType="text/plain">
        #             7504912.93758151 -764109.175074507,7750849.82379226 -22141.8611641468,8561828.42371234 -897195.923493867,7724946.16844165 -602984.014261927  # noqa
        #        </ns0:ComplexData>
        # </ns0:Data>
        # nspath('Data', ns=WPS_NAMESPACE)
        complex_data_element = element.find(
            nspath('ComplexData', ns=getNamespace(element)))
        if complex_data_element is not None:
            self.dataType = "ComplexData"

    def _parseLiteralData(self, element, literalElementName):
        """
        Method to parse the LiteralData element.
        """

        # <LiteralData>
        #    <ows:DataType ows:reference="xs:string" xmlns:ows="http://www.opengis.net/ows/1.1" />
        #    <ows:AllowedValues xmlns:ows="http://www.opengis.net/ows/1.1">
        #        <ows:Value>COMMA</ows:Value>
        #        <ows:Value>TAB</ows:Value>
        #        <ows:Value>SPACE</ows:Value>
        #    </ows:AllowedValues>
        #    <DefaultValue>COMMA</DefaultValue>
        # </LiteralData>

        # <LiteralData>
        #     <ows:DataType ows:reference="xs:anyURI" xmlns:ows="http://www.opengis.net/ows/1.1" />
        #     <ows:AnyValue xmlns:ows="http://www.opengis.net/ows/1.1" />
        # </LiteralData>

        literal_data_element = element.find(literalElementName)

        if literal_data_element is not None:
            self.dataType = 'LiteralData'
            for sub_element in literal_data_element:
                subns = getNamespace(sub_element)
                if sub_element.tag.endswith('DataType'):
                    self.dataType = sub_element.text
                    if not self.dataType:
                        reference = sub_element.get(nspath("reference", ns=subns))
                        # backward search of first non-alpha character (:, #, /, etc.)
                        pos = len(reference) - 1
                        while pos >= 0 and reference[pos].isalpha():
                            pos -= 1
                        # obtain substring after found non-alpha character position
                        self.dataType = reference[pos + 1:]

            for sub_element in literal_data_element:

                subns = getNamespace(sub_element)

                if sub_element.tag.endswith('DefaultValue'):
                    self.defaultValue = getTypedValue(
                        self.dataType, sub_element.text)

                if sub_element.tag.endswith('AllowedValues'):
                    for value in sub_element.findall(nspath('Value', ns=subns)):
                        self.allowedValues.append(
                            getTypedValue(self.dataType, value.text))
                elif sub_element.tag.endswith('AnyValue'):
                    self.anyValue = True

    def _parseComplexData(self, element, complexDataElementName):
        """
        Method to parse a ComplexData or ComplexOutput element.
        """

        # <ComplexData maximumMegabytes="200">
        #     <Default>
        #         <Format>
        #            <MimeType>text/xml</MimeType>
        #            <Encoding>UTF-8</Encoding>
        #            <Schema>http://schemas.opengis.net/gml/2.0.0/feature.xsd</Schema>
        #        </Format>
        #    </Default>
        #    <Supported>
        #        <Format>
        #            <MimeType>text/xml</MimeType>
        #            <Encoding>UTF-8</Encoding>
        #            <Schema>http://schemas.opengis.net/gml/2.0.0/feature.xsd</Schema>
        #        </Format>
        #        <Format>
        #            <MimeType>text/xml</MimeType>
        #            <Encoding>UTF-8</Encoding>
        #            <Schema>http://schemas.opengis.net/gml/2.1.1/feature.xsd</Schema>
        #        </Format>
        #    </Supported>
        # </ComplexData>
        # OR
        # <ComplexOutput defaultEncoding="UTF-8" defaultFormat="text/XML" defaultSchema="NONE">
        #     <SupportedComplexData>
        #         <Format>text/XML</Format>
        #         <Encoding>UTF-8</Encoding>
        #         <Schema>NONE</Schema>
        #     </SupportedComplexData>
        # </ComplexOutput>

        complex_data_element = element.find(complexDataElementName)
        if complex_data_element is not None:
            self.dataType = "ComplexData"

            for supported_comlexdata_element in\
                    complex_data_element.findall('SupportedComplexData'):
                self.supportedValues.append(
                    ComplexData(
                        mimeType=testXMLValue(
                            supported_comlexdata_element.find('Format')),
                        encoding=testXMLValue(
                            supported_comlexdata_element.find('Encoding')),
                        schema=testXMLValue(
                            supported_comlexdata_element.find('Schema'))
                    )
                )

            # Optional attribute 'maximumMegabytes' can be omitted entirely from 'ComplexData'
            # but in the case it is made available, report it in the format for reference.
            # This could be used by a file validation step to be respected during process execution
            # but this check should be done on the server side, so it is only informative here.
            max_mb = complex_data_element.attrib.get("maximumMegabytes")
            for format_element in\
                    complex_data_element.findall('Supported/Format'):
                self.supportedValues.append(
                    ComplexData(
                        mimeType=testXMLValue(format_element.find('MimeType')),
                        encoding=testXMLValue(format_element.find('Encoding')),
                        schema=testXMLValue(format_element.find('Schema')),
                        maximumMegaBytes=max_mb
                    )
                )

            default_format_element = complex_data_element.find('Default/Format')
            if default_format_element is not None:
                self.defaultValue = ComplexData(
                    mimeType=testXMLValue(
                        default_format_element.find('MimeType')),
                    encoding=testXMLValue(
                        default_format_element.find('Encoding')),
                    schema=testXMLValue(default_format_element.find('Schema')),
                    maximumMegaBytes=max_mb
                )

    def _parseBoundingBoxData(self, element, bboxElementName):
        """
        Method to parse the BoundingBoxData element.
        """

        # <BoundingBoxData>
        #   <Default>
        #     <CRS>epsg:4326</CRS>
        #   </Default>
        #   <Supported>
        #     <CRS>epsg:4326</CRS>
        #   </Supported>
        # </BoundingBoxData>
        #
        # OR
        #
        # <BoundingBoxOutput>
        #   <Default>
        #     <CRS>epsg:4326</CRS>
        #   </Default>
        #   <Supported>
        #     <CRS>epsg:4326</CRS>
        #   </Supported>
        # </BoundingBoxOutput>

        bbox_data_element = element.find(bboxElementName)
        if bbox_data_element is not None:
            self.dataType = 'BoundingBoxData'

            for bbox_element in bbox_data_element.findall('Supported/CRS'):
                self.supportedValues.append(bbox_element.text)

            default_bbox_element = bbox_data_element.find('Default/CRS')
            if default_bbox_element is not None:
                self.defaultValue = default_bbox_element.text


class Input(InputOutput):
    """
    Class that represents a WPS process input.
    """

    def __init__(self, inputElement):

        # superclass initializer
        super(Input, self).__init__(inputElement)

        # <Input maxOccurs="1" minOccurs="0">
        # OR
        # <MinimumOccurs>1</MinimumOccurs>
        self.minOccurs = -1
        if inputElement.get("minOccurs") is not None:
            self.minOccurs = int(inputElement.get("minOccurs"))
        if inputElement.find('MinimumOccurs') is not None:
            self.minOccurs = int(
                testXMLValue(inputElement.find('MinimumOccurs')))
        self.maxOccurs = -1
        if inputElement.get("maxOccurs") is not None:
            self.maxOccurs = int(inputElement.get("maxOccurs"))
        if inputElement.find('MaximumOccurs') is not None:
            self.maxOccurs = int(
                testXMLValue(inputElement.find('MaximumOccurs')))

        # <LiteralData>
        self._parseLiteralData(inputElement, 'LiteralData')

        # <ComplexData>
        self._parseComplexData(inputElement, 'ComplexData')

        # <BoundingBoxData>
        self._parseBoundingBoxData(inputElement, 'BoundingBoxData')


class Output(InputOutput):

    """
    Class that represents a WPS process output.
    """

    def __init__(self, outputElement):

        # superclass initializer
        super(Output, self).__init__(outputElement)

        self.reference = None
        self.mimeType = None
        self.data = []
        self.fileName = None
        self.filePath = None

        # extract wps namespace from outputElement itself
        wpsns = getNamespace(outputElement)

        # <ns:Reference encoding="UTF-8" mimeType="text/csv"
        # href="http://cida.usgs.gov/climate/gdp/process/RetrieveResultServlet?id=1318528582026OUTPUT.601bb3d0-547f-4eab-8642-7c7d2834459e"
        # />
        referenceElement = outputElement.find(nspath('Reference', ns=wpsns))
        if referenceElement is not None:
            # extract xlink namespace
            xlinkns = get_namespaces()['xlink']
            xlink_href = '{{{}}}href'.format(xlinkns)

            if xlink_href in list(referenceElement.keys()):
                self.reference = referenceElement.get(xlink_href)
            else:
                self.reference = referenceElement.get('href')
            self.mimeType = referenceElement.get('mimeType')

        # <LiteralOutput>
        self._parseLiteralData(outputElement, 'LiteralOutput')

        # <ComplexData> or <ComplexOutput>
        self._parseComplexData(outputElement, 'ComplexOutput')

        # <BoundingBoxOutput>
        self._parseBoundingBoxData(outputElement, 'BoundingBoxOutput')

        # <Data>
        # <ns0:Data>
        #        <ns0:ComplexData mimeType="text/plain">
        #             7504912.93758151 -764109.175074507,7750849.82379226 -22141.8611641468,8561828.42371234 -897195.923493867,7724946.16844165 -602984.014261927  # noqa
        #        </ns0:ComplexData>
        # </ns0:Data>
        # OR:
        # <ns0:Data>
        #     <ns0:ComplexData encoding="UTF-8" mimeType="text/xml" schema="http://schemas.opengis.net/gml/2.1.2/feature.xsd">  # noqa
        #         <ns3:FeatureCollection xsi:schemaLocation="http://ogr.maptools.org/ output_0n7ij9D.xsd" xmlns:ns3="http://ogr.maptools.org/">  # noqa
        #             <gml:boundedBy xmlns:gml="http://www.opengis.net/gml">
        #                 <gml:Box>
        #                     <gml:coord><gml:X>-960123.1421801626</gml:X><gml:Y>4665723.56559387</gml:Y></gml:coord>
        #                     <gml:coord><gml:X>-101288.6510608822</gml:X><gml:Y>5108200.011823481</gml:Y></gml:coord>
        #                 </gml:Box>
        #            </gml:boundedBy>
        #            <gml:featureMember xmlns:gml="http://www.opengis.net/gml">
        #                <ns3:output fid="F0">
        #                    <ns3:geometryProperty><gml:LineString><gml:coordinates>-960123.142180162365548,4665723.565593870356679,0 -960123.142180162365548,4665723.565593870356679,0 -960123.142180162598379,4665723.565593870356679,0 -960123.142180162598379,4665723.565593870356679,0 -711230.141176006174646,4710278.48552671354264,0 -711230.141176006174646,4710278.48552671354264,0 -623656.677859728806652,4848552.374973464757204,0 -623656.677859728806652,4848552.374973464757204,0   # noqa -410100.337491964863148,4923834.82589447684586,0 -410100.337491964863148,4923834.82589447684586,0 -101288.651060882242746,5108200.011823480948806,0 -101288.651060882242746,5108200.011823480948806,0 -101288.651060882257298,5108200.011823480948806,0 -101288.651060882257298,5108200.011823480948806,0</gml:coordinates></gml:LineString></ns3:geometryProperty>
        #                    <ns3:cat>1</ns3:cat>
        #                    <ns3:id>1</ns3:id>
        #                    <ns3:fcat>0</ns3:fcat>
        #                    <ns3:tcat>0</ns3:tcat>
        #                    <ns3:sp>0</ns3:sp>
        #                    <ns3:cost>1002619.181</ns3:cost>
        #                    <ns3:fdist>0</ns3:fdist>
        #                    <ns3:tdist>0</ns3:tdist>
        #                </ns3:output>
        #            </gml:featureMember>
        #        </ns3:FeatureCollection>
        #     </ns0:ComplexData>
        # </ns0:Data>
        #
        #
        # OWS BoundingBox:
        #
        # <wps:Data>
        #   <ows:BoundingBox crs="EPSG:4326" dimensions="2">
        #     <ows:LowerCorner>0.0 -90.0</ows:LowerCorner>
        #     <ows:UpperCorner>180.0 90.0</ows:UpperCorner>
        #   </ows:BoundingBox>
        # </wps:Data>
        #
        dataElement = outputElement.find(nspath('Data', ns=wpsns))
        if dataElement is not None:
            complexDataElement = dataElement.find(
                nspath('ComplexData', ns=wpsns))
            if complexDataElement is not None:
                self.dataType = "ComplexData"
                self.mimeType = complexDataElement.get('mimeType')
                if complexDataElement.text is not None and complexDataElement.text.strip() != '':
                    self.data.append(complexDataElement.text.strip())
                for child in complexDataElement:
                    self.data.append(etree.tostring(child))
            literalDataElement = dataElement.find(
                nspath('LiteralData', ns=wpsns))
            if literalDataElement is not None:
                self.dataType = literalDataElement.get('dataType')
                if literalDataElement.text is not None and literalDataElement.text.strip() != '':
                    self.data.append(literalDataElement.text.strip())
            bboxDataElement = dataElement.find(nspath('BoundingBoxData', ns=namespaces['wps']))
            if bboxDataElement is not None:
                self.dataType = "BoundingBoxData"
                bbox = BoundingBox(bboxDataElement)
                if bbox:
                    self.data.append(bbox)

    def retrieveData(self, username=None, password=None, headers=None, verify=True, cert=None):
        """
        Method to retrieve data from server-side reference:
        returns "" if the reference is not known.

        :param username: credentials to access the remote WPS server
        :param password: credentials to access the remote WPS server
        """
        url = self.reference
        if url is None:
            return b''

        # a) 'http://cida.usgs.gov/climate/gdp/process/RetrieveResultServlet?id=1318528582026OUTPUT.601bb3d0-547f-4eab-8642-7c7d2834459e'  # noqa
        # b) 'http://rsg.pml.ac.uk/wps/wpsoutputs/outputImage-11294Bd6l2a.tif'
        LOGGER.info('Output URL=%s' % url)

        # Extract output filepath from base URL
        self.fileName = url.split('/')[-1]

        # The link is a local file.
        # Useful when running local tests during development.
        if url.startswith("file://"):
            fn = url[7:]
            # If on Windows and path starts with a '/', remove it
            if os.name == 'nt' and fn.startswith('/'):
                fn = fn[1:]
            with open(fn) as f:
                return f.read()

        if '?' in url:
            spliturl = url.split('?')
            # Extract output filepath from URL query string
            self.fileName = spliturl[1].split('=')[1]

            u = openURL(spliturl[0], spliturl[
                        1], method='Get', username=username, password=password,
                        headers=headers, verify=verify, cert=cert)
        else:
            u = openURL(
                url, '', method='Get', username=username, password=password,
                headers=headers, verify=verify, cert=cert)

        return u.read()

    def writeToDisk(self, path=None, username=None, password=None,
                    headers=None, verify=True, cert=None):
        """
        Method to write an output of a WPS process to disk:
        it either retrieves the referenced file from the server, or write out the content of response embedded output.

        :param path: optional path to the output file, otherwise a file will be created in the local directory
                  with the name assigned by the server,
        :param username: credentials to access the remote WPS server
        :param password: credentials to access the remote WPS server
        """
        # Check if ExecuteResponse contains reference to server-side output
        content = self.retrieveData(username, password, headers=headers, verify=verify, cert=cert)

        # ExecuteResponse contain embedded output
        if content == "" and len(self.data) > 0:
            self.fileName = self.identifier
            for data in self.data:
                content = content + data

        # write out content
        if content != "":
            if self.fileName == "":
                self.fileName = self.identifier
            self.filePath = path + self.fileName
            out = open(self.filePath, 'wb')
            out.write(content)
            out.close()
            LOGGER.info('Output written to file: %s' % self.filePath)


class WPSException:
    """
    Class representing an exception raised by a WPS.
    """
    def __init__(self, root):
        self.code = root.attrib.get("exceptionCode", None)
        self.locator = root.attrib.get("locator", None)
        textEl = root.find(nspath('ExceptionText', ns=getNamespace(root)))
        if textEl is not None:
            self.text = textEl.text
        else:
            self.text = ""


class Metadata(object):
    """Initialize an OWS Metadata construct"""
    def __init__(self, elem, namespace=DEFAULT_OWS_NAMESPACE):
        self.url = None
        self.title = None
        self.role = None

        if elem is not None:
            urlattrib = elem.attrib.get(nspath('href', XLINK_NAMESPACE))
            if urlattrib is not None:
                self.url = testXMLValue(urlattrib, True)
            titleattrib = elem.attrib.get(nspath('title', XLINK_NAMESPACE))
            if titleattrib is not None:
                self.title = testXMLValue(titleattrib, True)
            roleattrib = elem.attrib.get(nspath('role', XLINK_NAMESPACE))
            if roleattrib is not None:
                self.role = testXMLValue(roleattrib, True)


class Process(object):
    """
    Class that represents a WPS process.
    """
    def __init__(self, elem):
        """ Initialization method extracts all available metadata from an XML document (passed in as etree object) """

        # <ns0:ProcessDescriptions service="WPS" version="1.0.0"
        #   xsi:schemaLocation="http://www.opengis.net/wps/1.0.0 http://schemas.opengis.net/wps/1.0.0/wpsDescribeProcess_response.xsd"  # noqa
        #   xml:lang="en-US" xmlns:ns0="http://www.opengis.net/wps/1.0.0" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">  # noqa
        # OR:
        # <ns0:Process ns0:processVersion="1.0.0">
        self._root = elem

        # when process is instantiated from GetCapabilities, elem is 'wps:Process'          => wpsns='wps'
        # when process is instantiated from DescribeProcess, elem is 'ProcessDescription'   => wpsns=''
        wpsns = getNamespace(elem) or n.get_namespace('wps')

        def get_bool_attribute(elem, attribute):
            property = elem.get(attribute, '').lower()
            if property == 'true':
                value = True
            elif property == 'false':
                value = False
            else:
                value = None
            return value

        # <ProcessDescription statusSupported="true" storeSupported="true" ns0:processVersion="1.0.0">
        #   because version attribute is always 'wps:processVersion', wpsns with both variants must be tested
        #   in order to handle both elem in wpsns='wps' and wpsns='' cases
        self.processVersion = elem.get(nspath('processVersion', ns=wpsns)) or elem.get(nspath('processVersion'))
        self.statusSupported = get_bool_attribute(elem, "statusSupported")
        self.storeSupported = get_bool_attribute(elem, "storeSupported")
        self.identifier = None
        self.title = None
        self.abstract = None
        self.metadata = []

        for child in elem:

            # this element's namespace
            # TODO: ns not used
            # ns = getNamespace(child)

            # <ows:Identifier xmlns:ows="http://www.opengis.net/ows/1.1">
            #   gov.usgs.cida.gdp.wps.algorithm.FeatureWeightedGridStatisticsAlgorithm
            # </ows:Identifier>
            if child.tag.endswith('Identifier'):
                self.identifier = testXMLValue(child)

            # <ows:Title xmlns:ows="http://www.opengis.net/ows/1.1">
            #   Feature Weighted Grid Statistics
            # </ows:Title>
            elif child.tag.endswith('Title'):
                self.title = testXMLValue(child)

            # <ows:Abstract xmlns:ows="http://www.opengis.net/ows/1.1">
            #   This algorithm generates area weighted statistics of a gridded dataset for
            #   a set of vector polygon features. Using the bounding-box that encloses
            #   the feature data and the time range, if provided, a subset of the gridded dataset
            #   is requested from the remote gridded data server.
            #   Polygon representations are generated for cells in the retrieved grid.
            #   The polygon grid-cell representations are then projected to the feature data
            #   coordinate reference system. The grid-cells are used to calculate per grid-cell
            #   feature coverage fractions. Area-weighted statistics are then calculated for each feature
            #   using the grid values and fractions as weights. If the gridded dataset has a time range
            #   the last step is repeated for each time step within the time range or all time steps
            #   if a time range was not supplied.
            # </ows:Abstract>
            elif child.tag.endswith('Abstract'):
                self.abstract = testXMLValue(child)

            # <ows:Metadata xlink:title="Documentation" xlink:href="http://emu.readthedocs.org/en/latest/"/>
            elif child.tag.endswith('Metadata'):
                self.metadata.append(Metadata(child))

        LOGGER.debug(self)

        # <DataInputs>
        self.dataInputs = []
        for inputElement in elem.findall('DataInputs/Input'):
            self.dataInputs.append(Input(inputElement))
            LOGGER.debug(self.dataInputs[-1])

        # <ProcessOutputs>
        self.processOutputs = []
        for outputElement in elem.findall('ProcessOutputs/Output'):
            self.processOutputs.append(Output(outputElement))
            LOGGER.debug(self.processOutputs[-1])

    def __str__(self):
        return "WPS Process: {}, title={}".format(self.identifier or '', self.title or '')

    def __repr__(self):
        return "<owslib.wps.Process {}>".format(self.identifier or '')


class BoundingBoxDataInput(object):
    """
    Data input class for ``wps:BoundingBoxData``.

    :param list data: Coordinates of lower and upper corner. Example [10, 50, 20, 60]
    with lower corner=[10, 50] and upper corner=[20, 60].
    :param str crs: Name of coordinate reference system. Default: "epsg:4326".
    """
    def __init__(self, data, crs=None, dimensions=2):
        if isinstance(data, list):
            self.data = data
        else:
            # convenience method for string input
            self.data = [float(number) for number in data.split(',')]
        self.lower_corner = (self.data[0], self.data[1])
        self.upper_corner = (self.data[2], self.data[3])
        self.dimensions = dimensions
        self.crs = crs or 'epsg:4326'

    def get_xml(self):
        """
        Method that returns the object data as an XML snippet,
        to be inserted into the WPS request document sent to the server.
        """
        '''
        <wps:Data>
            <wps:BoundingBoxData crs="EPSG:4326" dimenstions="2">
                <ows:LowerCorner>51.9 7.0</ows:LowerCorner>
                <ows:UpperCorner>53.0 8.0</ows:UpperCorner>
            </wps:BoundingBoxData>
        </wps:Data>
        '''
        data_el = etree.Element(nspath_eval('wps:Data', namespaces))
        attrib = dict()
        if self.crs:
            attrib['crs'] = self.crs
        if self.dimensions:
            attrib['dimensions'] = str(self.dimensions)
        bbox_el = etree.SubElement(
            data_el, nspath_eval('wps:BoundingBoxData', namespaces), attrib=attrib)
        lc_el = etree.SubElement(
            bbox_el, nspath_eval('ows:LowerCorner', namespaces))
        lc_el.text = "{0[0]} {0[1]}".format(self.lower_corner)
        uc_el = etree.SubElement(
            bbox_el, nspath_eval('ows:UpperCorner', namespaces))
        uc_el.text = "{0[0]} {0[1]}".format(self.upper_corner)
        return data_el


class ComplexDataInput(IComplexDataInput, ComplexData):

    def __init__(self, value, mimeType=None, encoding=None, schema=None):
        super(ComplexDataInput, self).__init__(
            mimeType=mimeType, encoding=encoding, schema=schema)
        self.value = value

    def getXml(self):
        if is_reference(self.value):
            return self.complexDataAsReference()
        else:
            return self.complexDataRaw()

    def complexDataAsReference(self):
        """
           <wps:Reference xlink:href="http://somewhere/test.xml"/>
        """
        attrib = {nspath_eval("xlink:href", namespaces): self.value}
        if self.encoding:
            attrib['encoding'] = self.encoding
        if self.schema:
            attrib['schema'] = self.schema
        if self.mimeType:
            attrib['mimeType'] = self.mimeType
        refElement = etree.Element(nspath_eval('wps:Reference', namespaces), attrib)
        return refElement

    def complexDataRaw(self):
        '''
            <wps:Data>
                <wps:ComplexData mimeType="text/xml" encoding="UTF-8"
                    schema="http://schemas.opengis.net/gml/3.1.1/base/feature.xsd">
                </wps:ComplexData>
            </wps:Data>
        '''
        dataElement = etree.Element(nspath_eval('wps:Data', namespaces))

        attrib = dict()
        if self.encoding:
            attrib['encoding'] = self.encoding
        if self.schema:
            attrib['schema'] = self.schema
        if self.mimeType:
            attrib['mimeType'] = self.mimeType
        complexDataElement = etree.SubElement(
            dataElement, nspath_eval('wps:ComplexData', namespaces), attrib=attrib)
        complexDataElement.text = self.value
        return dataElement


class FeatureCollection(IComplexDataInput):

    '''
    Base class to represent a Feature Collection used as input to a WPS request.
    The method getXml() is invoked by the WPS execute() method to build the WPS request.
    All subclasses must implement the getXml() method to provide their specific XML.

    Implements IComplexDataInput.
    '''

    def __init__(self):
        pass

    def getXml(self):
        raise NotImplementedError


class WFSFeatureCollection(FeatureCollection):

    '''
    FeatureCollection specified by a WFS query.
    All subclasses must implement the getQuery() method to provide the specific query portion of the XML.
    '''
    def __init__(self, wfsUrl, wfsQuery, wfsMethod=None):
        '''
        wfsUrl: the WFS service URL
                example: wfsUrl = "http://igsarm-cida-gdp2.er.usgs.gov:8082/geoserver/wfs"
        wfsQuery : a WFS query instance
        '''
        self.url = wfsUrl
        self.query = wfsQuery
        self.method = wfsMethod

    #    <wps:Reference xlink:href="http://igsarm-cida-gdp2.er.usgs.gov:8082/geoserver/wfs">
    #      <wps:Body>
    #        <wfs:GetFeature xmlns:wfs="http://www.opengis.net/wfs" xmlns:ogc="http://www.opengis.net/ogc" xmlns:gml="http://www.opengis.net/gml" service="WFS" version="1.1.0" outputFormat="text/xml; subtype=gml/3.1.1" xsi:schemaLocation="http://www.opengis.net/wfs ../wfs/1.1.0/WFS.xsd">  # noqa
    #            .......
    #        </wfs:GetFeature>
    #      </wps:Body>
    #   </wps:Reference>
    def getXml(self):
        root = etree.Element(nspath_eval('wps:Reference', namespaces),
                             attrib={nspath_eval("xlink:href", namespaces): self.url})
        if self.method:
            root.attrib['method'] = self.method
        bodyElement = etree.SubElement(
            root, nspath_eval('wps:Body', namespaces))
        getFeatureElement = etree.SubElement(
            bodyElement, nspath_eval('wfs:GetFeature', namespaces),
            attrib={"service": "WFS",
                    "version": "1.1.0",
                    "outputFormat": "text/xml; subtype=gml/3.1.1",
                    nspath_eval("xsi:schemaLocation", namespaces): "%s %s" % (namespaces['wfs'], WFS_SCHEMA_LOCATION)})

        #            <wfs:Query typeName="sample:CONUS_States">
        #                <wfs:PropertyName>the_geom</wfs:PropertyName>
        #                <wfs:PropertyName>STATE</wfs:PropertyName>
        #                <ogc:Filter>
        #                    <ogc:GmlObjectId gml:id="CONUS_States.508"/>
        #                </ogc:Filter>
        #            </wfs:Query>
        getFeatureElement.append(self.query.getXml())

        return root


class WFSQuery(IComplexDataInput):

    '''
    Class representing a WFS query, for insertion into a WFSFeatureCollection instance.

    Implements IComplexDataInput.
    '''

    def __init__(self, typeName, propertyNames=[], filters=[]):
        self.typeName = typeName
        self.propertyNames = propertyNames
        self.filters = filters

    def getXml(self):

        #            <wfs:Query typeName="sample:CONUS_States">
        #                <wfs:PropertyName>the_geom</wfs:PropertyName>
        #                <wfs:PropertyName>STATE</wfs:PropertyName>
        #                <ogc:Filter>
        #                    <ogc:GmlObjectId gml:id="CONUS_States.508"/>
        #                </ogc:Filter>
        #            </wfs:Query>

        queryElement = etree.Element(
            nspath_eval('wfs:Query', namespaces), attrib={"typeName": self.typeName})
        for propertyName in self.propertyNames:
            propertyNameElement = etree.SubElement(
                queryElement, nspath_eval('wfs:PropertyName', namespaces))
            propertyNameElement.text = propertyName
        if len(self.filters) > 0:
            filterElement = etree.SubElement(
                queryElement, nspath_eval('ogc:Filter', namespaces))
            for filter in self.filters:
                # gmlObjectIdElement
                etree.SubElement(
                    filterElement, nspath_eval('ogc:GmlObjectId', namespaces),
                    attrib={nspath_eval('gml:id', namespaces): filter})
        return queryElement


class GMLMultiPolygonFeatureCollection(FeatureCollection):
    '''
    Class that represents a FeatureCollection defined as a GML multi-polygon.
    '''
    def __init__(self, polygons):
        '''
        Initializer accepts an array of polygons, where each polygon is an array of (lat,lon) tuples.
        Example: polygons = [ [(-102.8184, 39.5273), (-102.8184, 37.418), (-101.2363, 37.418), (-101.2363, 39.5273), (-102.8184, 39.5273)],  # noqa
                              [(-92.8184, 39.5273), (-92.8184, 37.418), (-91.2363, 37.418), (-91.2363, 39.5273), (-92.8184, 39.5273)] ]
        '''
        self.polygons = polygons

    def getXml(self):
        '''
            <wps:Data>
                <wps:ComplexData mimeType="text/xml" encoding="UTF-8"
                    schema="http://schemas.opengis.net/gml/3.1.1/base/feature.xsd">
                    <gml:featureMembers xmlns:ogc="http://www.opengis.net/ogc"
                        xmlns:draw="gov.usgs.cida.gdp.draw" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                        xmlns:ows="http://www.opengis.net/ows" xmlns:gml="http://www.opengis.net/gml"
                        xmlns:xlink="http://www.w3.org/1999/xlink"
                        xsi:schemaLocation="gov.usgs.cida.gdp.draw http://cida.usgs.gov/climate/derivative/xsd/draw.xsd">  # noqa
                        <gml:box gml:id="box.1">
                            <gml:the_geom>
                                <gml:MultiPolygon srsDimension="2"
                                    srsName="http://www.opengis.net/gml/srs/epsg.xml#4326">
                                    <gml:polygonMember>
                                        <gml:Polygon>
                                            <gml:exterior>
                                                <gml:LinearRing>
                                                    <gml:posList>-102.8184 39.5273 -102.8184 37.418 -101.2363 37.418 -101.2363 39.5273 -102.8184 39.5273</gml:posList>
                                                </gml:LinearRing>
                                            </gml:exterior>
                                        </gml:Polygon>
                                    </gml:polygonMember>
                                </gml:MultiPolygon>
                            </gml:the_geom>
                            <gml:ID>0</gml:ID>
                        </gml:box>
                    </gml:featureMembers>
                </wps:ComplexData>
            </wps:Data>
        '''
        dataElement = etree.Element(nspath_eval('wps:Data', namespaces))
        complexDataElement = etree.SubElement(
            dataElement, nspath_eval('wps:ComplexData', namespaces),
            attrib={"mimeType": "text/xml", "schema": GML_SCHEMA_LOCATION})
        featureMembersElement = etree.SubElement(
            complexDataElement, nspath_eval('gml:featureMembers', namespaces),
            attrib={nspath_eval("xsi:schemaLocation", namespaces): "%s %s" % (DRAW_NAMESPACE, DRAW_SCHEMA_LOCATION)})
        boxElement = etree.SubElement(featureMembersElement, nspath_eval(
            'gml:box', namespaces), attrib={nspath_eval("gml:id", namespaces): "box.1"})
        geomElement = etree.SubElement(
            boxElement, nspath_eval('gml:the_geom', namespaces))
        multiPolygonElement = etree.SubElement(
            geomElement, nspath_eval('gml:MultiPolygon', namespaces),
            attrib={"srsDimension": "2", "srsName": "http://www.opengis.net/gml/srs/epsg.xml#4326"})
        for polygon in self.polygons:
            polygonMemberElement = etree.SubElement(
                multiPolygonElement, nspath_eval('gml:polygonMember', namespaces))
            polygonElement = etree.SubElement(
                polygonMemberElement, nspath_eval('gml:Polygon', namespaces))
            exteriorElement = etree.SubElement(
                polygonElement, nspath_eval('gml:exterior', namespaces))
            linearRingElement = etree.SubElement(
                exteriorElement, nspath_eval('gml:LinearRing', namespaces))
            posListElement = etree.SubElement(
                linearRingElement, nspath_eval('gml:posList', namespaces))
            posListElement.text = ' '.join(
                ["%s %s" % (x, y) for x, y in polygon[:]])

        idElement = etree.SubElement(
            boxElement, nspath_eval('gml:ID', namespaces))
        idElement.text = "0"
        return dataElement


def monitorExecution(execution, sleepSecs=3, download=False, filepath=None):
    '''
    Convenience method to monitor the status of a WPS execution till it completes (succesfully or not),
    and write the output to file after a succesfull job completion.

    :param execution: WPSExecution instance
    :param int sleepSecs: number of seconds to sleep in between check status invocations
    :param download: True to download the output when the process terminates, False otherwise
    :param filepath: optional path to output file (if downloaded=True), otherwise filepath
              will be inferred from response document

    '''
    while execution.isComplete() is False:
        execution.checkStatus(sleepSecs=sleepSecs)
        LOGGER.info('Execution status: %s' % execution.status)

    if execution.isSucceeded():
        if download:
            execution.getOutput(filepath=filepath)
        else:
            for output in execution.processOutputs:
                if output.reference is not None:
                    LOGGER.info('Output URL=%s' % output.reference)
    else:
        for ex in execution.errors:
            LOGGER.error('Error: code=%s, locator=%s, text=%s' %
                         (ex.code, ex.locator, ex.text))


def printValue(value):
    '''
    Utility method to format a value for printing.
    '''

    # ComplexData type
    if isinstance(value, ComplexData):
        return "mimeType=%s, encoding=%s, schema=%s" % (value.mimeType, value.encoding, value.schema)
    # other type
    else:
        return value


def printInputOutput(value, indent=''):
    '''
    Utility method to inspect an input/output element.
    '''

    # InputOutput fields
    print(('{} identifier={}, title={}, abstract={}, data type={}'.format(
          indent, value.identifier, value.title, value.abstract, value.dataType)))
    for val in value.allowedValues:
        print(('{} Allowed Value: {}'.format(indent, printValue(val))))
    if value.anyValue:
        print(' Any value allowed')
    for val in value.supportedValues:
        print(('{} Supported Value: {}'.format(indent, printValue(val))))
    print(('{} Default Value: {} '.format(indent, printValue(value.defaultValue))))

    # Input fields
    if isinstance(value, Input):
        print(('{} minOccurs={}, maxOccurs={}'.format(
              indent, value.minOccurs, value.maxOccurs)))

    # Output fields
    if isinstance(value, Output):
        print(('{} reference={}, mimeType={}'.format(
              indent, value.reference, value.mimeType)))
        for datum in value.data:
            print(('{} Data Value: {}'.format(indent, printValue(datum))))


class Languages(object):
    """Initialize a WPS Languages construct"""
    def __init__(self, infoset):
        self._root = infoset
        self.default = None
        self.supported = []

        for element in self._root:
            if element.tag.endswith('Default'):
                self.default = testXMLValue(element[0])
            elif element.tag.endswith('Supported'):
                for child in element:
                    self.supported.append(testXMLValue(child))

    def __repr__(self):
        return "<owslib.wps.Languages default='{}' supported={}>".format(self.default, self.supported)



# =============================================================================
# Copyright (c) 2023 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

"""
API for A9 OpenSearch (description and query syntax).

https://github.com/dewitt/opensearch

Supports version 1.1 core as well as geo, time and parameters extensions.
"""

import logging
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

from owslib.etree import etree
from owslib.owscontext.atom import decode_atomxml
from owslib.util import Authentication, http_get, nspath_eval, testXMLValue

LOGGER = logging.getLogger(__name__)

namespaces = {
    'os': 'http://a9.com/-/spec/opensearch/1.1/',
    'geo': 'http://a9.com/-/spec/opensearch/extensions/geo/1.0/',
    'parameters': 'http://a9.com/-/spec/opensearch/extensions/parameters/1.0/',  # noqa
    'time': 'http://a9.com/-/spec/opensearch/extensions/time/1.0/'
}


class OpenSearch(object):
    """ OpenSearch request class """
    def __init__(self, url, lang='en-US', version='1.1', timeout=10, xml=None,
                 username=None, password=None, auth=None, headers=None):
        """

        Initialize an OpenSearch client

        Parameters
        ----------

        - url: the URL of the OpenSearch endpoint
        - lang: the language (default is 'en-US')
        - version: version (default is '2.0.2')
        - timeout: timeout in seconds
        - xml: string of OpenSearch Description Document
        - username: username for HTTP basic authentication
        - password: password for HTTP basic authentication
        - auth: instance of owslib.util.Authentication
        - headers: HTTP headers to send with requests
        """

        if auth:
            if username:
                auth.username = username
            if password:
                auth.password = password

        self.url = url
        self.lang = lang
        self.version = version
        self.timeout = timeout
        self.auth = auth or Authentication(username, password)
        self.headers = headers

        if xml is not None:
            LOGGER.debug('Loading from XML stream')
            self._exml = etree.fromstring(xml)
        else:
            LOGGER.debug('Loading from URL')
            self.request = self.url
            response = http_get(self.request, timeout=self.timeout)
            self._exml = etree.fromstring(response.content)

        self.description = Description(self._exml)

    def search(self, type_: str, **kwargs) -> dict:
        """
        Invoke an OpenSearch search

        :param type_: media type of search
        :param kwargs: `dict` of key value pairs. When an OpenSearch
                       description document has URL Parameters, kwargs keys must
                       match accordingly.  When no OpenSearch URL Parameters
                       are defined, To pass kwargs keys which are OpenSearch
                       parameters such as `geo:time` or `eo:orbitNumber`, pass
                       the keys as `geo_time` or `eo_orbitNumber`.

        :returns: dict of response
        """

        if type_ not in self.description.urls.keys():
            msg = 'Invalid URL type'
            raise RuntimeError(msg)

        template = self.description.urls[type_]['template']

        if self.description.urls[type_]['parameters']:
            LOGGER.debug('Validating kwargs against defined parameters')
            for key, value in kwargs.items():
                if key not in self.description.urls[type_]['parameters']:
                    msg = f'parameter {key} not found'
                    LOGGER.debug(msg)
                    raise RuntimeError(msg)
                if 'options' in self.description.urls[type_]['parameters'][key]:
                    LOGGER.debug('Validating against parameter options')
                    if value not in self.description.urls[type_]['parameters'][key]['options']:
                        msg = f"{value} not in {self.description.urls[type_]['parameters'][key]['options']}"
                        LOGGER.debug(msg)
                        raise RuntimeError(msg)

                LOGGER.debug(f'Setting parameter {key} in URL template')

                template = template_replace_token(
                    template,
                    self.description.urls[type_]['parameters'][key]['value'],
                    value)

        else:
            LOGGER.debug('Best effort against no defined parameters')
            for key, value in kwargs.items():
                template = template_replace_token(
                    template,
                    key.replace('_', ':'),
                    value)

        response = http_get(prune_url(template), timeout=self.timeout)

        if 'json' in type_:
            LOGGER.debug('Returning dict of JSON response')
            response = response.json()
        elif 'atom' in type_:
            LOGGER.debug('Returning dict of Atom response')
            response = decode_atomxml(response.content)
        else:
            LOGGER.debug('Unknown/unsupported response, returning as is')

        return response


class Description:
    def __init__(self, md):

        self.urls = {}

        LOGGER.debug('Parsing Description')

        val = md.find(nspath_eval('os:ShortName', namespaces))
        self.shortname = testXMLValue(val)

        val = md.find(nspath_eval('os:LongName', namespaces))
        self.longname = testXMLValue(val)

        val = md.find(nspath_eval('os:Description', namespaces))
        self.description = testXMLValue(val)

        val = md.find(nspath_eval('os:Language', namespaces))
        self.language = testXMLValue(val)

        val = md.find(nspath_eval('os:InputEncoding', namespaces))
        self.inputencoding = testXMLValue(val)

        val = md.find(nspath_eval('os:OutputEncoding', namespaces))
        self.outputencoding = testXMLValue(val)

        val = md.find(nspath_eval('os:Tags', namespaces))
        self.tags = testXMLValue(val).split()

        val = md.find(nspath_eval('os:Contact', namespaces))
        self.contact = testXMLValue(val)

        val = md.find(nspath_eval('os:Developer', namespaces))
        self.developer = testXMLValue(val)

        val = md.find(nspath_eval('os:Attribution', namespaces))
        self.attribution = testXMLValue(val)

        val = md.find(nspath_eval('os:SyndicationRight', namespaces))
        self.syndicationright = testXMLValue(val)

        val = md.find(nspath_eval('os:AdultContent', namespaces))
        self.adultcontent = testXMLValue(val)

        for u in md.findall(nspath_eval('os:Url', namespaces)):
            url_type = u.attrib.get('type')

            url_def = {
                'rel': u.attrib.get('rel'),
                'template': u.attrib.get('template'),
                'parameters': {}
            }
            for p in u.findall(nspath_eval('parameters:Parameter', namespaces)):
                p_name = p.attrib.get('name')
                p_def = {
                    'pattern': p.attrib.get('pattern'),
                    'title': p.attrib.get('title'),
                    'value': p.attrib.get('value')
                }

                options = [o.attrib.get('value') for o in p.findall(nspath_eval('parameters:Option', namespaces))]
                if len(options) > 0:
                    p_def['options'] = options

                url_def['parameters'][p_name] = p_def

            self.urls[url_type] = url_def


def template_replace_token(template: str, token: str, value: str) -> str:
    """
    Helper function to replace OpenSearch token in a URL template

    :param template: URL template
    :param token: token to replace
    :param value: value to replace token with

    :returns: updated URL template
    """

    token2 = token.replace('}', '?}')

    if token2 in template:
        LOGGER.debug('Replacing optional token')
        token_to_replace = token2
    else:
        LOGGER.debug('Replacing required token')
        token_to_replace = token

    LOGGER.debug(f'Token to replace: {token_to_replace}')

    LOGGER.debug(f'Template: {template}')
    template2 = template.replace(token_to_replace, value)
    LOGGER.debug(f'Template (replaced): {template2}')

    return template2


def prune_url(url: str) -> str:
    """
    Helper function to prune URL of unused template parameters

    https://stackoverflow.com/a/2506477

    :param url: URL

    :returns: updated URL without unused template parameters
    """

    query_params_out = {}

    url_parts = list(urlparse(url))

    query_params_in = dict(parse_qsl(url_parts[4]))

    for key, value in query_params_in.items():
        if '{' not in value and '}' not in value:
            query_params_out[key] = value

    url_parts[4] = urlencode(query_params_out)

    return urlunparse(url_parts)



# -*- coding: utf-8 -*-
# =============================================================================
# Authors : Alexander Kmoch <allixender@gmail.com>
#
# =============================================================================

"""
place for some constants to avoid circular imports
"""

# seen in wms.py
from urllib.parse import urlparse

from datetime import datetime
from dateutil import parser

"""
The spec reference uri is of course difference for each encoding,
so holding it in the context container object is maybe a bit silly,
thus I use a generic spec reference and and only encode at serialisation time
"""

GENERIC_OWCSPEC_URL = "http://www.opengis.net/spec/owc-generic/1.0/req"
ATOM_OWCSPEC_URL = "http://www.opengis.net/spec/owc-atom/1.0/req"
GEOJSON_OWCSPEC_URL = "http://www.opengis.net/spec/owc-geojson/1.0/req"

"""
supported geojson profiles
"""

SUPPORTED_GEOJSON_PROFILES = [GEOJSON_OWCSPEC_URL + "/core"]

"""
those are for compliance testing when parsing geojson owc input
"""

SUPPORTED_GEOJSON_OFFERING_EXTENSIONS = [
    GEOJSON_OWCSPEC_URL + "/wms",
    GEOJSON_OWCSPEC_URL + "/wfs",
    GEOJSON_OWCSPEC_URL + "/wcs",
    GEOJSON_OWCSPEC_URL + "/wps",
    GEOJSON_OWCSPEC_URL + "/csw",
    GEOJSON_OWCSPEC_URL + "/geotiff",
    GEOJSON_OWCSPEC_URL + "/sos"
]

# FIXME are the geosjson and atom offering codes ok,
# because ATOM offering codes are different (spec vs conf vs req)

ATOM_OFFERING_CODES = [
    'http://www.opengis.net/spec/owc/1.0/conf/atom/gml',
    'http://www.opengis.net/spec/owc/1.0/req/atom/wms'
    'http://www.opengis.net/spec/owc-atom/1.0/req/gml',
    'http://www.opengis.net/spec/owc-atom/1.0/req/csw',
]


def encodedspecurl_to_genericspecurl(encodedspecurl, genericspecurl):
    parsed = urlparse(encodedspecurl)
    speccode = "/" + parsed.path.split("/").last.trim
    return genericspecurl + speccode


def genericspecurl_to_encodedspecurl(genericspecurl, encodedspecurl):
    parsed = urlparse(genericspecurl)
    speccode = parsed.path.split("/").last.trim
    return encodedspecurl + speccode


class TimeIntervalFormat(object):
    """
    little helper to have time intervals
    """

    def __init__(self,
                 start,
                 end=None):
        """
        constructor:

        :param start: datetime
        :param end: datetime
        """
        self.start = start
        self.end = end

    def __str__(self):
        if self.end is None:
            return self.start.isoformat()
        else:
            return self.start.isoformat() + "/" + self.end.isoformat()

    def to_dict(self):
        """
        dict representation of object, for simple object comparison
        :return: dict
        """
        return {
            "start": None if self.start is None else self.start.isoformat(),
            "end": None if self.start is None else self.start.isoformat()
        }

    @classmethod
    def from_string(cls, date_str):
        if date_str is None:
            return None
        try:
            date_arr = date_str.split("/")
            if len(date_arr) > 1:
                start_dt = parser.parse(date_arr[0])
                end_dt = parser.parse(date_arr[1])
                return TimeIntervalFormat(start_dt, end_dt)
            else:
                single_dt = parser.parse(date_str)
                return TimeIntervalFormat(single_dt)
        except Exception:
            raise ValueError("Error parsing datetime string: %s" % date_str)


def skip_nulls(dict_obj):
    """
    drops key/val pairs where js value is null, not needed in the JSON
    :param o: needs to be dict
    :return:
    """
    reduced = {k: v for k, v in list(dict_obj.items()) if v is not None}
    return reduced


def skip_nulls_rec(dict_obj):
    """
    removes dict key/val pairs recursively where value is None,
    not needed/wanted(?) in the JSON
    :param o: needs to be dict
    :return: the trimed dict, or exceptionally the value if it wasn't a dict
    """
    if not isinstance(dict_obj, dict):
        return dict_obj
    else:
        result = {}
        for k, v in list(dict_obj.items()):
            if v is None:
                pass
            else:
                if isinstance(v, dict):
                    tmp = skip_nulls_rec(v)
                    result.update({k: tmp})
                elif isinstance(v, list):
                    tmp = [skip_nulls_rec(o) for o in v]
                    result.update({k: tmp})
                else:
                    result.update({k: v})
        return result


def extract_p(path, dict_obj, default):
    """
    try to extract dict value in key path, if key error provide default
    :param path: the nested dict key path, separated by '.'
    (therefore no dots in key names allowed)
    :param dict_obj: the dictinary object from which to extract
    :param default: a default return value if key error
    :return: extracted value
    """
    if dict_obj is None:
        return default

    keys = path.split('.')
    tmp_iter = dict_obj
    for key in keys:
        try:
            # dict.get() might make KeyError exception unnecessary
            tmp_iter = tmp_iter.get(key, default)
        except KeyError:
            return default
    return tmp_iter


def build_from_xp(path, dict_obj, build_class, default):
    """
    try to build class instance from extracted path, else return default
    :param path: path: the nested dict key path, separated by '.'
    (therefore no dots in key names allowed)
    :param dict_obj: the dictinary object from which to extract
    :param build_class: the class name from which to build
    :param default: default return value
    :return: ideally the inquired class instance, else default
    """
    xp = extract_p(path, dict_obj, default)
    if xp is None:
        return default
    elif xp == default:
        return default
    else:
        return build_class.from_dict(xp)


def is_empty(dict_obj):
    """
    query if a dict is empty

    :param dict_obj: the to be tested dictionary
    :return: True, if it is empty, False if not empty
    """
    if isinstance(dict_obj, dict):
        if len(list(dict_obj.items())) <= 0:
            return True
        else:
            switch = True
            for k, v in list(dict_obj.items()):
                if v is None:
                    pass
                else:
                    switch = False
            return switch
    else:
        return False


def try_int(num_string):
    """
    short hand cast to number

    :param num_string:
    :return: int or None
    """
    if num_string is not None:
        try:
            return int(num_string)
        except ValueError:
            pass
    return None


def try_float(num_string):
    """
    short hand cast to number

    :param num_string:
    :return: float or None
    """
    if num_string is not None:
        try:
            return float(num_string)
        except ValueError:
            pass
    return None



# -*- coding: ISO-8859-15 -*-
# =============================================================================
# Copyright (c) 2009 Tom Kralidis
#
# Authors : Tom Kralidis <tomkralidis@gmail.com>
#           Angelos Tzotsos <tzotsos@gmail.com>
#
# Contact email: tomkralidis@gmail.com
# =============================================================================

# flake8: noqa: E501

""" ISO metadata parser """

from owslib.etree import etree
from owslib import util
from owslib.namespaces import Namespaces


# default variables
def get_namespaces():
    n = Namespaces()
    ns = n.get_namespaces(["gco", "gfc", "gmd", "gmi", "gml", "gml32", "gmx", "gts", "srv", "xlink"])
    ns[None] = n.get_namespace("gmd")
    return ns


namespaces = get_namespaces()

def testFirstCharOrAnchor(md,xpath):
    """ function which checks if the first matching element is either charstring or anchor, if anchor, returns {name, url}, else {name} """
    r = {'name': '', 'url': ''}
    val = md.find(util.nspath_eval(xpath+'/gco:CharacterString', namespaces))
    r['name'] = util.testXMLValue(val)
    if r['name'] in [None,'']:    
        val = md.find(util.nspath_eval(xpath+'/gmx:Anchor', namespaces))
        if val is not None:
            r['name'] = util.testXMLValue(val)
            r['url'] = val.attrib.get(util.nspath_eval('xlink:href', namespaces))
    return r

def testAllCharOrAnchor(md,xpath,aslist=True):
    """ 
    function which checks for each matching element if it is charstring or anchor, returns {name, uri}
    the aslist parameter indicates to return 2 lists or single list of objects
    """
    ro = []
    for i in md.findall(util.nspath_eval(xpath+'/gco:CharacterString', namespaces)):
        r = {'name': '', 'url': ''}
        r['name'] = util.testXMLValue(i)
        ro.append(r)
    for i in md.findall(util.nspath_eval(xpath+'/gmx:Anchor', namespaces)):
        r = {'name': '', 'url': ''}
        if i is not None:
            r['name'] = util.testXMLValue(i)
            r['url'] = i.attrib.get(util.nspath_eval('xlink:href', namespaces)) 
            ro.append(r)
    if aslist:
        return ro
    else:
        return {'name': [i.get('name', '') for i in ro],
                'url': [i.get('url', '') for i in ro]}



class MD_Metadata(object):
    """ Process gmd:MD_Metadata """
    def __init__(self, md=None):

        if md is None:
            self.md = None
            self.xml = None
            self.identifier = None
            self.parentidentifier = None
            self.language = None
            self.dataseturi = None
            self.languagecode = None
            self.datestamp = None
            self.charset = None
            self.hierarchy = None
            self.contact = []
            self.datetimestamp = None
            self.stdname = None
            self.stdver = None
            self.locales = []
            self.referencesystem = None
            self.contentinfo = None
            self.identification = []
            self.contentinfo = []
            self.distribution = None
            self.dataquality = None
            self.acquisition = None
        else:
            self.md = md
            if hasattr(md, 'getroot'):  # standalone document
                self.xml = etree.tostring(md.getroot())
            else:  # part of a larger document
                self.xml = etree.tostring(md)

            val = md.find(util.nspath_eval('gmd:fileIdentifier/gco:CharacterString', namespaces))
            self.identifier = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:parentIdentifier/gco:CharacterString', namespaces))
            self.parentidentifier = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:language/gco:CharacterString', namespaces))
            self.language = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:dataSetURI/gco:CharacterString', namespaces))
            self.dataseturi = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:language/gmd:LanguageCode', namespaces))
            self.languagecode = util.testXMLAttribute(val, 'codeListValue')

            val = md.find(util.nspath_eval('gmd:dateStamp/gco:Date', namespaces))
            self.datestamp = util.testXMLValue(val)

            if not self.datestamp:
                val = md.find(util.nspath_eval('gmd:dateStamp/gco:DateTime', namespaces))
                self.datestamp = util.testXMLValue(val)

            self.charset = _testCodeListValue(md.find(
                util.nspath_eval('gmd:characterSet/gmd:MD_CharacterSetCode', namespaces)))

            self.hierarchy = _testCodeListValue(md.find(
                util.nspath_eval('gmd:hierarchyLevel/gmd:MD_ScopeCode', namespaces)))

            self.contact = []
            for i in md.findall(util.nspath_eval('gmd:contact/gmd:CI_ResponsibleParty', namespaces)):
                o = CI_ResponsibleParty(i)
                self.contact.append(o)

            val = md.find(util.nspath_eval('gmd:dateStamp/gco:DateTime', namespaces))
            self.datetimestamp = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:metadataStandardName/gco:CharacterString', namespaces))
            self.stdname = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:metadataStandardVersion/gco:CharacterString', namespaces))
            self.stdver = util.testXMLValue(val)

            self.locales = []
            for i in md.findall(util.nspath_eval('gmd:locale/gmd:PT_Locale', namespaces)):
                self.locales.append(PT_Locale(i))

            val = md.find(util.nspath_eval('gmd:referenceSystemInfo/gmd:MD_ReferenceSystem', namespaces))
            if val is not None:
                self.referencesystem = MD_ReferenceSystem(val)
            else:
                self.referencesystem = None

            self.identification = []

            for idinfo in md.findall(util.nspath_eval('gmd:identificationInfo', namespaces)):
                if len(idinfo) > 0:
                    val = list(idinfo)[0]
                    tagval = util.xmltag_split(val.tag)
                    if tagval == 'MD_DataIdentification':
                        self.identification.append(MD_DataIdentification(val, 'dataset'))
                    elif tagval == 'MD_ServiceIdentification':
                        self.identification.append(MD_DataIdentification(val, 'service'))
                    elif tagval == 'SV_ServiceIdentification':
                        self.identification.append(SV_ServiceIdentification(val))

            self.contentinfo = []
            for contentinfo in md.findall(
                    util.nspath_eval('gmd:contentInfo/gmd:MD_FeatureCatalogueDescription', namespaces)):
                self.contentinfo.append(MD_FeatureCatalogueDescription(contentinfo))
            for contentinfo in md.findall(
                    util.nspath_eval('gmd:contentInfo/gmd:MD_ImageDescription', namespaces)):
                self.contentinfo.append(MD_ImageDescription(contentinfo))

            val = md.find(util.nspath_eval('gmd:distributionInfo/gmd:MD_Distribution', namespaces))

            if val is not None:
                self.distribution = MD_Distribution(val)
            else:
                self.distribution = None

            val = md.find(util.nspath_eval('gmd:dataQualityInfo/gmd:DQ_DataQuality', namespaces))
            if val is not None:
                self.dataquality = DQ_DataQuality(val)
            else:
                self.dataquality = None

            val = md.find(util.nspath_eval('gmi:acquisitionInformation/gmi:MI_AcquisitionInformation', namespaces))
            if val is not None:
                self.acquisition = MI_AcquisitionInformation(val)
            else:
                self.acquisition = None

    def get_all_contacts(self):
        """get all contacts in document"""

        contacts = []

        for ii in self.identification:
            for iic in ii.contact:
                contacts.append(iic)

            for ct in ['creator', 'publisher', 'contributor']:
                iict = getattr(ii, ct)
                if iict:
                    contacts.append(iict)

        return list(filter(None, contacts))

    def get_default_locale(self):
        """ get default gmd:PT_Locale based on gmd:language """

        for loc in self.locales:
            if loc.languagecode == self.language:
                return loc
        return None


class PT_Locale(object):
    """ process PT_Locale """

    def __init__(self, md=None):
        if md is None:
            self.id = None
            self.languagecode = None
            self.charset = None
        else:
            self.id = md.attrib.get('id')
            self.languagecode = None
            languagecode = md.find(util.nspath_eval('gmd:languageCode/gmd:LanguageCode', namespaces))
            if languagecode is not None:
                self.languagecode = languagecode.attrib.get('codeListValue')
            self.charset = None
            charset = md.find(util.nspath_eval('gmd:characterEncoding/gmd:MD_CharacterSetCode', namespaces))
            if charset is not None:
                self.charset = charset.attrib.get('codeListValue')


class CI_Date(object):
    """ process CI_Date """
    def __init__(self, md=None):
        if md is None:
            self.date = None
            self.type = None
        else:
            val = md.find(util.nspath_eval('gmd:date/gco:Date', namespaces))
            if val is not None:
                self.date = util.testXMLValue(val)
            else:
                val = md.find(util.nspath_eval('gmd:date/gco:DateTime', namespaces))
                if val is not None:
                    self.date = util.testXMLValue(val)
                else:
                    self.date = None

            val = md.find(util.nspath_eval('gmd:dateType/gmd:CI_DateTypeCode', namespaces))
            self.type = _testCodeListValue(val)


class CI_ResponsibleParty(object):
    """ process CI_ResponsibleParty """
    def __init__(self, md=None):

        if md is None:
            self.name = None
            self.name_url = None
            self.organization = None
            self.organization_url = None
            self.position = None
            self.phone = None
            self.fax = None
            self.address = None
            self.city = None
            self.region = None
            self.postcode = None
            self.country = None
            self.email = None
            self.onlineresource = None
            self.role = None
        else:

            frm = testFirstCharOrAnchor(md,'gmd:individualName')
            self.name = frm['name']
            self.name_url = frm['url']

            frm = testFirstCharOrAnchor(md,'gmd:organisationName')
            self.organization = frm['name']
            self.organization_url = frm['url']

            val = md.find(util.nspath_eval('gmd:positionName/gco:CharacterString', namespaces))
            self.position = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:phone/gmd:CI_Telephone/gmd:voice/gco:CharacterString', namespaces))

            self.phone = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:phone/gmd:CI_Telephone/gmd:facsimile/gco:CharacterString',
                namespaces))
            self.fax = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:deliveryPoint/gco:CharacterString',
                namespaces))
            self.address = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:city/gco:CharacterString', namespaces))
            self.city = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:administrativeArea/gco:CharacterString',
                namespaces))
            self.region = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:postalCode/gco:CharacterString',
                namespaces))
            self.postcode = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:country/gco:CharacterString',
                namespaces))
            self.country = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:address/gmd:CI_Address/gmd:electronicMailAddress/gco:CharacterString',
                namespaces))
            self.email = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:contactInfo/gmd:CI_Contact/gmd:onlineResource/gmd:CI_OnlineResource', namespaces))
            if val is not None:
                self.onlineresource = CI_OnlineResource(val)
            else:
                self.onlineresource = None

            self.role = _testCodeListValue(md.find(util.nspath_eval('gmd:role/gmd:CI_RoleCode', namespaces)))


class Keyword(object):
    """
    Class for complex keywords, with labels and URLs
    """
    def __init__(self, kw=None):
        if kw is None:
            self.name = None
            self.url = None
        else:
            self.name = util.testXMLValue(kw)
            self.url = kw.attrib.get(util.nspath_eval('xlink:href', namespaces))


class MD_Keywords(object):
    """
    Class for the metadata MD_Keywords element
    """
    def __init__(self, md=None):
        if md is None:
            self.keywords = []
            self.type = None
            self.thesaurus = None
            self.kwdtype_codeList = 'http://standards.iso.org/ittf/PubliclyAvailableStandards/ISO_19139_Schemas/resources/codelist/gmxCodelists.xml#MD_KeywordTypeCode'
        else:
            self.keywords = []
            val = md.findall(util.nspath_eval('gmd:keyword/gco:CharacterString', namespaces))
            if len(val) == 0:
                val = md.findall(util.nspath_eval('gmd:keyword/gmx:Anchor', namespaces))
            for word in val:
                self.keywords.append(Keyword(word))
            self.type = None
            val = md.find(util.nspath_eval('gmd:type/gmd:MD_KeywordTypeCode', namespaces))
            self.type = util.testXMLAttribute(val, 'codeListValue')

            self.thesaurus = None
            val = md.find(util.nspath_eval('gmd:thesaurusName/gmd:CI_Citation', namespaces))
            if val is not None:
                self.thesaurus = {}

                title = val.find(util.nspath_eval('gmd:title/gco:CharacterString', namespaces))
                self.thesaurus['title'] = util.testXMLValue(title)
                self.thesaurus['url'] = None

                if self.thesaurus['title'] is None:  # try gmx:Anchor
                    t = val.find(util.nspath_eval('gmd:title/gmx:Anchor', namespaces))
                    if t is not None:
                        self.thesaurus['title'] = util.testXMLValue(t)
                        self.thesaurus['url'] = t.attrib.get(util.nspath_eval('xlink:href', namespaces))

                date_ = val.find(util.nspath_eval('gmd:date/gmd:CI_Date/gmd:date/gco:Date', namespaces))
                self.thesaurus['date'] = util.testXMLValue(date_)

                datetype = val.find(
                    util.nspath_eval('gmd:date/gmd:CI_Date/gmd:dateType/gmd:CI_DateTypeCode', namespaces))
                self.thesaurus['datetype'] = util.testXMLAttribute(datetype, 'codeListValue')


class MD_DataIdentification(object):
    """ process MD_DataIdentification """
    def __init__(self, md=None, identtype=None):
        if md is None:
            self.identtype = None
            self.title = None
            self.alternatetitle = None
            self.aggregationinfo = None
            self.uricode = []
            self.uricodespace = []
            self.date = []
            self.datetype = []
            self.uselimitation = []
            self.uselimitation_url = []
            self.accessconstraints = []
            self.classification = []
            self.otherconstraints = []
            self.securityconstraints = []
            self.useconstraints = []
            self.denominators = []
            self.distance = []
            self.uom = []
            self.resourcelanguage = []
            self.resourcelanguagecode = []
            self.creator = []
            self.publisher = []
            self.contributor = []
            self.edition = None
            self.abstract = None
            self.abstract_url = None
            self.purpose = None
            self.status = None
            self.graphicoverview = []
            self.contact = []
            self.keywords = []
            self.topiccategory = []
            self.supplementalinformation = None
            self.extent = None
            self.bbox = None
            self.temporalextent_start = None
            self.temporalextent_end = None
            self.spatialrepresentationtype = []
        else:
            self.identtype = identtype
            val = md.find(util.nspath_eval('gmd:citation/gmd:CI_Citation/gmd:title/gco:CharacterString', namespaces))
            self.title = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'gmd:citation/gmd:CI_Citation/gmd:alternateTitle/gco:CharacterString', namespaces))
            self.alternatetitle = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:aggregationInfo', namespaces))
            self.aggregationinfo = util.testXMLValue(val)

            self.uricode = []
            _values = md.findall(util.nspath_eval(
                'gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:MD_Identifier/gmd:code/gco:CharacterString',
                namespaces))
            _values += md.findall(util.nspath_eval(
                'gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:RS_Identifier/gmd:code/gco:CharacterString',
                namespaces))
            for i in _values:
                val = util.testXMLValue(i)
                if val not in [None,'']:
                    self.uricode.append(val)
            
            _values = md.findall(util.nspath_eval(
                'gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:MD_Identifier/gmd:code/gmx:Anchor',
                namespaces))
            _values += md.findall(util.nspath_eval(
                'gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:RS_Identifier/gmd:code/gmx:Anchor',
                namespaces))
            for i in _values:
                val = util.testXMLValue(i)
                val1 = i.attrib.get(util.nspath_eval('xlink:href', namespaces))
                if val1 not in [None,'']:
                    self.uricode.append(val1)
                elif val not in [None,'']:
                    self.uricode.append(val)
                    

            self.uricodespace = []
            for i in md.findall(util.nspath_eval(
                    'gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:RS_Identifier/gmd:codeSpace/gco:CharacterString',
                    namespaces)):
                val = util.testXMLValue(i)
                if val not in [None,'']:
                    self.uricodespace.append(val)
            for i in md.findall(util.nspath_eval(
                    'gmd:citation/gmd:CI_Citation/gmd:identifier/gmd:RS_Identifier/gmd:codeSpace/gmx:Anchor',
                    namespaces)):
                val = util.testXMLValue(i)
                val1 = i.attrib.get(util.nspath_eval('xlink:href', namespaces))
                if val1 not in [None,'']:
                    self.uricode.append(val1)
                elif val not in [None,'']:
                    self.uricode.append(val)


            self.date = []
            self.datetype = []

            for i in md.findall(util.nspath_eval('gmd:citation/gmd:CI_Citation/gmd:date/gmd:CI_Date', namespaces)):
                self.date.append(CI_Date(i))

            self.uselimitation = []
            self.uselimitation_url = []
            _values = md.findall(util.nspath_eval(
                'gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:useLimitation/gco:CharacterString', namespaces))
            _values += md.findall(util.nspath_eval(
                'gmd:resourceConstraints/gmd:MD_Constraints/gmd:useLimitation/gco:CharacterString', namespaces))
            for i in _values:
                val = util.testXMLValue(i)
                if val is not None:
                    self.uselimitation.append(val)

            _values = md.findall(util.nspath_eval(
                'gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:useLimitation/gmx:Anchor', namespaces))
            _values += md.findall(util.nspath_eval(
                'gmd:resourceConstraints/gmd:MD_Constraints/gmd:useLimitation/gmx:Anchor', namespaces))
            for i in _values:
                val = util.testXMLValue(i)
                val1 = i.attrib.get(util.nspath_eval('xlink:href', namespaces))

                if val is not None:
                    self.uselimitation.append(val)
                    self.uselimitation_url.append(val1)

            self.accessconstraints = []
            for i in md.findall(util.nspath_eval(
                    'gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:accessConstraints/gmd:MD_RestrictionCode',
                    namespaces)):
                val = _testCodeListValue(i)
                if val is not None:
                    self.accessconstraints.append(val)

            self.classification = []
            for i in md.findall(util.nspath_eval(
                    'gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:accessConstraints/gmd:MD_ClassificationCode',
                    namespaces)):
                val = _testCodeListValue(i)
                if val is not None:
                    self.classification.append(val)

            ocs = testAllCharOrAnchor(md,'gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:otherConstraints', False)
            self.otherconstraints = ocs['name']
            self.otherconstraints_url = ocs['url']

            self.securityconstraints = []
            for i in md.findall(util.nspath_eval(
                    'gmd:resourceConstraints/gmd:MD_SecurityConstraints/gmd:classification/gmd:MD_ClassificationCode',
                    namespaces)):
                val = _testCodeListValue(i)
                if val is not None:
                    self.securityconstraints.append(val)

            self.useconstraints = []
            for i in md.findall(util.nspath_eval(
                    'gmd:resourceConstraints/gmd:MD_LegalConstraints/gmd:useConstraints/gmd:MD_RestrictionCode',
                    namespaces)):
                val = _testCodeListValue(i)
                if val is not None:
                    self.useconstraints.append(val)

            self.denominators = []
            for i in md.findall(util.nspath_eval(
                    'gmd:spatialResolution/gmd:MD_Resolution/gmd:equivalentScale/gmd:MD_RepresentativeFraction/gmd:denominator/gco:Integer',
                    namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.denominators.append(val)

            self.distance = []
            self.uom = []
            for i in md.findall(util.nspath_eval(
                    'gmd:spatialResolution/gmd:MD_Resolution/gmd:distance/gco:Distance', namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.distance.append(val)
                self.uom.append(i.get("uom"))

            self.resourcelanguagecode = []
            for i in md.findall(util.nspath_eval('gmd:language/gmd:LanguageCode', namespaces)):
                val = _testCodeListValue(i)
                if val is not None:
                    self.resourcelanguagecode.append(val)

            self.resourcelanguage = []
            for i in md.findall(util.nspath_eval('gmd:language/gco:CharacterString', namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.resourcelanguage.append(val)

            self.creator = []
            self.publisher = []
            self.contributor = []
            for val in md.findall(util.nspath_eval('gmd:pointOfContact/gmd:CI_ResponsibleParty', namespaces)):
                role = val.find(util.nspath_eval('gmd:role/gmd:CI_RoleCode', namespaces))
                if role is not None:
                    clv = _testCodeListValue(role)
                    rp = CI_ResponsibleParty(val)
                    if clv == 'originator':
                        self.creator.append(rp)
                    elif clv == 'publisher':
                        self.publisher.append(rp)
                    elif clv == 'author':
                        self.contributor.append(rp)

            val = md.find(util.nspath_eval('gmd:citation/gmd:CI_Citation/gmd:edition/gco:CharacterString', namespaces))
            self.edition = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:abstract/gco:CharacterString', namespaces))
            self.abstract = util.testXMLValue(val)

            val = md.find(util.nspath_eval('gmd:abstract/gmx:Anchor', namespaces))

            self.abstract_url = None
            if val is not None:
                self.abstract = util.testXMLValue(val)
                self.abstract_url = val.attrib.get(util.nspath_eval('xlink:href', namespaces))

            val = md.find(util.nspath_eval('gmd:purpose/gco:CharacterString', namespaces))
            self.purpose = util.testXMLValue(val)

            self.status = _testCodeListValue(md.find(util.nspath_eval('gmd:status/gmd:MD_ProgressCode', namespaces)))

            self.graphicoverview = []
            for val in md.findall(util.nspath_eval(
                    'gmd:graphicOverview/gmd:MD_BrowseGraphic/gmd:fileName/gco:CharacterString', namespaces)):
                if val is not None:
                    val2 = util.testXMLValue(val)
                    if val2 is not None:
                        self.graphicoverview.append(val2)

            self.contact = []
            for i in md.findall(util.nspath_eval('gmd:pointOfContact/gmd:CI_ResponsibleParty', namespaces)):
                o = CI_ResponsibleParty(i)
                self.contact.append(o)

            self.spatialrepresentationtype = []
            for val in md.findall(util.nspath_eval(
                    'gmd:spatialRepresentationType/gmd:MD_SpatialRepresentationTypeCode', namespaces)):
                val = util.testXMLAttribute(val, 'codeListValue')
                if val:
                    self.spatialrepresentationtype.append(val)

            self.keywords = []
            for mdkw in md.findall(util.nspath_eval('gmd:descriptiveKeywords/gmd:MD_Keywords', namespaces)):
                self.keywords.append(MD_Keywords(mdkw))

            self.topiccategory = []
            for i in md.findall(util.nspath_eval('gmd:topicCategory/gmd:MD_TopicCategoryCode', namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.topiccategory.append(val)

            val = md.find(util.nspath_eval('gmd:supplementalInformation/gco:CharacterString', namespaces))
            self.supplementalinformation = util.testXMLValue(val)

            # There may be multiple geographicElement, create an extent
            # from the one containing either an EX_GeographicBoundingBox or EX_BoundingPolygon.
            # The schema also specifies an EX_GeographicDescription. This is not implemented yet.
            val = None
            val2 = None
            val3 = None
            extents = md.findall(util.nspath_eval('gmd:extent', namespaces))
            extents.extend(md.findall(util.nspath_eval('srv:extent', namespaces)))
            for extent in extents:
                if val is None:
                    for e in extent.findall(util.nspath_eval('gmd:EX_Extent/gmd:geographicElement', namespaces)):
                        if e.find(util.nspath_eval('gmd:EX_GeographicBoundingBox', namespaces)) is not None or \
                                e.find(util.nspath_eval('gmd:EX_BoundingPolygon', namespaces)) is not None:
                            val = e
                            break
                    self.extent = EX_Extent(val)
                    self.bbox = self.extent.boundingBox  # for backwards compatibility

                if val2 is None:
                    val2 = extent.find(util.nspath_eval(
                        'gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:beginPosition',
                        namespaces))
                    if val2 is None:
                        val2 = extent.find(util.nspath_eval(
                            'gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml32:TimePeriod/gml32:beginPosition',
                            namespaces))
                    self.temporalextent_start = util.testXMLValue(val2)

                if val3 is None:
                    val3 = extent.find(util.nspath_eval(
                        'gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml:TimePeriod/gml:endPosition',
                        namespaces))
                    if val3 is None:
                        val3 = extent.find(util.nspath_eval(
                            'gmd:EX_Extent/gmd:temporalElement/gmd:EX_TemporalExtent/gmd:extent/gml32:TimePeriod/gml32:endPosition',
                            namespaces))
                    self.temporalextent_end = util.testXMLValue(val3)


class MD_Distributor(object):
    """ process MD_Distributor """
    def __init__(self, md=None):
        if md is None:
            self.contact = None
            self.online = []
        else:
            self.contact = None
            val = md.find(util.nspath_eval(
                'gmd:MD_Distributor/gmd:distributorContact/gmd:CI_ResponsibleParty', namespaces))
            if val is not None:
                self.contact = CI_ResponsibleParty(val)

            self.online = []

            for ol in md.findall(util.nspath_eval(
                    'gmd:MD_Distributor/gmd:distributorTransferOptions/gmd:MD_DigitalTransferOptions/gmd:onLine/gmd:CI_OnlineResource',
                    namespaces)):
                self.online.append(CI_OnlineResource(ol))


class MD_Distribution(object):
    """ process MD_Distribution """
    def __init__(self, md=None):
        if md is None:
            self.format = None
            self.format_url = None
            self.version = None
            self.version_url = None
            self.specification = None
            self.specification_url = None
            self.distributor = []
            self.online = []
            pass
        else:
            frm = testFirstCharOrAnchor(md,'gmd:distributionFormat/gmd:MD_Format/gmd:name')
            self.format = frm['name']
            self.format_url = frm['url']
            
            vrs = testFirstCharOrAnchor(md,'gmd:distributionFormat/gmd:MD_Format/gmd:version')
            self.version = vrs['name']
            self.version_url = vrs['url']

            spc = testFirstCharOrAnchor(md,'gmd:distributionFormat/gmd:MD_Format/gmd:specification')
            self.specification = spc['name']
            self.specification_url = spc['url']

            self.distributor = []
            for dist in md.findall(util.nspath_eval('gmd:distributor', namespaces)):
                self.distributor.append(MD_Distributor(dist))

            self.online = []
            for ol in md.findall(util.nspath_eval(
                    'gmd:transferOptions/gmd:MD_DigitalTransferOptions/gmd:onLine/gmd:CI_OnlineResource',
                    namespaces)):
                self.online.append(CI_OnlineResource(ol))


class DQ_DataQuality(object):
    ''' process DQ_DataQuality'''
    def __init__(self, md=None):
        if md is None:
            self.conformancetitle = []
            self.conformancetitle_url = []
            self.conformancedate = []
            self.conformancedatetype = []
            self.conformancedegree = []
            self.lineage = None
            self.lineage_url = None
            self.specificationtitle = None
            self.specificationdate = []
        else:
            cts = testAllCharOrAnchor(md,'gmd:report/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:title', False)
            self.conformancetitle = cts['name']
            self.conformancetitle_url = cts['url']

            self.conformancedate = []
            for i in md.findall(util.nspath_eval(
                    'gmd:report/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:date/gmd:CI_Date/gmd:date/gco:Date',
                    namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.conformancedate.append(val)

            self.conformancedatetype = []
            for i in md.findall(util.nspath_eval(
                    'gmd:report/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:date/gmd:CI_Date/gmd:dateType/gmd:CI_DateTypeCode',
                    namespaces)):
                val = _testCodeListValue(i)
                if val is not None:
                    self.conformancedatetype.append(val)

            self.conformancedegree = []
            for i in md.findall(util.nspath_eval(
                    'gmd:report/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult/gmd:pass/gco:Boolean',
                    namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.conformancedegree.append(val)

            lng = testFirstCharOrAnchor(md, 'gmd:lineage/gmd:LI_Lineage/gmd:statement')
            self.lineage = lng['name']
            self.lineage_url = lng['url']

            val = md.find(util.nspath_eval(
                'gmd:report/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:title/gco:CharacterString',
                namespaces))
            self.specificationtitle = util.testXMLValue(val)

            self.specificationdate = []
            for i in md.findall(util.nspath_eval(
                    'gmd:report/gmd:DQ_DomainConsistency/gmd:result/gmd:DQ_ConformanceResult/gmd:specification/gmd:CI_Citation/gmd:date/gmd:CI_Date',
                    namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.specificationdate.append(val)


class SV_ServiceIdentification(MD_DataIdentification):
    """ process SV_ServiceIdentification """
    def __init__(self, md=None):

        super().__init__(md, 'service')

        if md is None:
            self.type = None
            self.version = None
            self.fees = None
            self.couplingtype = None
            self.operations = []
            self.operateson = []
        else:
            val = md.find(util.nspath_eval('srv:serviceType/gco:LocalName', namespaces))
            self.type = util.testXMLValue(val)

            val = md.find(util.nspath_eval('srv:serviceTypeVersion/gco:CharacterString', namespaces))
            self.version = util.testXMLValue(val)

            val = md.find(util.nspath_eval(
                'srv:accessProperties/gmd:MD_StandardOrderProcess/gmd:fees/gco:CharacterString', namespaces))
            self.fees = util.testXMLValue(val)

            self.couplingtype = _testCodeListValue(md.find(util.nspath_eval(
                'gmd:couplingType/gmd:SV_CouplingType', namespaces)))

            self.operations = []

            for i in md.findall(util.nspath_eval('srv:containsOperations', namespaces)):
                tmp = {}
                val = i.find(util.nspath_eval(
                    'srv:SV_OperationMetadata/srv:operationName/gco:CharacterString', namespaces))
                tmp['name'] = util.testXMLValue(val)
                tmp['dcplist'] = []
                for d in i.findall(util.nspath_eval('srv:SV_OperationMetadata/srv:DCP', namespaces)):
                    tmp2 = _testCodeListValue(d.find(util.nspath_eval('srv:DCPList', namespaces)))
                    tmp['dcplist'].append(tmp2)

                tmp['connectpoint'] = []

                for d in i.findall(util.nspath_eval('srv:SV_OperationMetadata/srv:connectPoint', namespaces)):
                    tmp3 = d.find(util.nspath_eval('gmd:CI_OnlineResource', namespaces))
                    tmp['connectpoint'].append(CI_OnlineResource(tmp3))
                self.operations.append(tmp)

            self.operateson = []

            for i in md.findall(util.nspath_eval('srv:operatesOn', namespaces)):
                tmp = {}
                tmp['uuidref'] = i.attrib.get('uuidref')
                tmp['href'] = i.attrib.get(util.nspath_eval('xlink:href', namespaces))
                tmp['title'] = i.attrib.get(util.nspath_eval('xlink:title', namespaces))
                self.operateson.append(tmp)


class CI_OnlineResource(object):
    """ process CI_OnlineResource """
    def __init__(self, md=None):
        if md is None:
            self.url = None
            self.protocol = None
            self.protocol_url = None
            self.name = None
            self.name_url = None
            self.description = None
            self.description_url = None
            self.function = None
            self.applicationprofile = None
            self.applicationprofile_url = None
        else:
            val = md.find(util.nspath_eval('gmd:linkage/gmd:URL', namespaces))
            self.url = util.testXMLValue(val)

            val = testFirstCharOrAnchor(md,'gmd:protocol')
            self.protocol = val['name']
            self.protocol_url = val['url']

            val = testFirstCharOrAnchor(md,'gmd:name')
            self.name = val['name']
            self.name_url = val['url']

            val = testFirstCharOrAnchor(md,'gmd:description')
            self.description = val['name']
            self.description_url = val['url']

            val = testFirstCharOrAnchor(md,'gmd:applicationProfile')
            self.applicationprofile = val['name']
            self.applicationprofile_url = val['url']

            val = md.find(util.nspath_eval('gmd:description/gco:CharacterString', namespaces))
            self.description = util.testXMLValue(val)

            self.function = _testCodeListValue(md.find(util.nspath_eval(
                'gmd:function/gmd:CI_OnLineFunctionCode', namespaces)))


class EX_GeographicBoundingBox(object):
    def __init__(self, md=None):
        if md is None:
            self.minx = None
            self.maxx = None
            self.miny = None
            self.maxy = None
        else:
            val = md.find(util.nspath_eval('gmd:westBoundLongitude/gco:Decimal', namespaces))
            self.minx = util.testXMLValue(val)
            val = md.find(util.nspath_eval('gmd:eastBoundLongitude/gco:Decimal', namespaces))
            self.maxx = util.testXMLValue(val)
            val = md.find(util.nspath_eval('gmd:southBoundLatitude/gco:Decimal', namespaces))
            self.miny = util.testXMLValue(val)
            val = md.find(util.nspath_eval('gmd:northBoundLatitude/gco:Decimal', namespaces))
            self.maxy = util.testXMLValue(val)


class EX_Polygon(object):
    def __init__(self, md=None):
        if md is None:
            self.exterior_ring = None
            self.interior_rings = []
        else:
            linear_ring = md.find(util.nspath_eval('gml32:Polygon/gml32:exterior/gml32:LinearRing', namespaces))
            if linear_ring is not None:
                self.exterior_ring = self._coordinates_for_ring(linear_ring)

            interior_ring_elements = md.findall(util.nspath_eval('gml32:Polygon/gml32:interior', namespaces))
            self.interior_rings = []
            for iring_element in interior_ring_elements:
                linear_ring = iring_element.find(util.nspath_eval('gml32:LinearRing', namespaces))
                self.interior_rings.append(self._coordinates_for_ring(linear_ring))

    def _coordinates_for_ring(self, linear_ring):
        coordinates = []
        positions = linear_ring.findall(util.nspath_eval('gml32:pos', namespaces))
        for pos in positions:
            tokens = pos.text.split()
            coords = tuple([float(t) for t in tokens])
            coordinates.append(coords)
        return coordinates


class EX_GeographicBoundingPolygon(object):
    def __init__(self, md=None):
        if md is None:
            self.is_extent = None
            self.polygons = []
        else:
            val = md.find(util.nspath_eval('gmd:extentTypeCode', namespaces))
            self.is_extent = util.testXMLValue(val)

            md_polygons = md.findall(util.nspath_eval('gmd:polygon', namespaces))

            self.polygons = []
            for val in md_polygons:
                self.polygons.append(EX_Polygon(val))


class EX_Extent(object):
    """ process EX_Extent """
    def __init__(self, md=None):
        if md is None:
            self.boundingBox = None
            self.boundingPolygon = None
            self.description_code = None
        else:
            self.boundingBox = None
            self.boundingPolygon = None

            if md is not None:
                bboxElement = md.find(util.nspath_eval('gmd:EX_GeographicBoundingBox', namespaces))
                if bboxElement is not None:
                    self.boundingBox = EX_GeographicBoundingBox(bboxElement)

                polygonElement = md.find(util.nspath_eval('gmd:EX_BoundingPolygon', namespaces))
                if polygonElement is not None:
                    self.boundingPolygon = EX_GeographicBoundingPolygon(polygonElement)

                val = md.find(util.nspath_eval(
                    'gmd:EX_GeographicDescription/gmd:geographicIdentifier/gmd:MD_Identifier/gmd:code/gco:CharacterString',
                    namespaces))
                self.description_code = util.testXMLValue(val)


class MD_ReferenceSystem(object):
    """ process MD_ReferenceSystem """
    def __init__(self, md=None):
        if md is None:
            self.code = None
            self.code_url = None
            self.codeSpace = None
            self.codeSpace_url = None
            self.version = None
            self.version_url = None
        else:
            val = testFirstCharOrAnchor(md,'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:code')
            self.code_url = val['url']      
            self.code = val['name']

            val = testFirstCharOrAnchor(md,'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:codeSpace')
            self.codeSpace_url = val['url']
            self.codeSpace = val['name']

            val = testFirstCharOrAnchor(md,'gmd:referenceSystemIdentifier/gmd:RS_Identifier/gmd:version')
            self.version_url = val['url']
            self.version = val['name']

def _testCodeListValue(elpath):
    """ get gco:CodeListValue_Type attribute, else get text content """
    if elpath is not None:  # try to get @codeListValue
        val = util.testXMLValue(elpath.attrib.get('codeListValue'), True)
        if val is not None:
            return val
        else:  # see if there is element text
            return util.testXMLValue(elpath)
    else:
        return None


class CodelistCatalogue(object):
    """ process CT_CodelistCatalogue """
    def __init__(self, ct):
        val = ct.find(util.nspath_eval('gmx:name/gco:CharacterString', namespaces))
        self.name = util.testXMLValue(val)
        val = ct.find(util.nspath_eval('gmx:scope/gco:CharacterString', namespaces))
        self.scope = util.testXMLValue(val)
        val = ct.find(util.nspath_eval('gmx:fieldOfApplication/gco:CharacterString', namespaces))
        self.fieldapp = util.testXMLValue(val)
        val = ct.find(util.nspath_eval('gmx:versionNumber/gco:CharacterString', namespaces))
        self.version = util.testXMLValue(val)
        val = ct.find(util.nspath_eval('gmx:versionDate/gco:Date', namespaces))
        self.date = util.testXMLValue(val)

        self.dictionaries = {}

        for i in ct.findall(util.nspath_eval('gmx:codelistItem/gmx:CodeListDictionary', namespaces)):
            id = i.attrib.get(util.nspath_eval('gml32:id', namespaces))
            self.dictionaries[id] = {}
            val = i.find(util.nspath_eval('gml32:description', namespaces))
            self.dictionaries[id]['description'] = util.testXMLValue(val)
            val = i.find(util.nspath_eval('gml32:identifier', namespaces))
            self.dictionaries[id]['identifier'] = util.testXMLValue(val)
            self.dictionaries[id]['entries'] = {}

            for j in i.findall(util.nspath_eval('gmx:codeEntry', namespaces)):
                id2 = j.find(util.nspath_eval('gmx:CodeDefinition', namespaces)).attrib.get(
                    util.nspath_eval('gml32:id', namespaces))
                self.dictionaries[id]['entries'][id2] = {}
                val = j.find(util.nspath_eval('gmx:CodeDefinition/gml32:description', namespaces))
                self.dictionaries[id]['entries'][id2]['description'] = util.testXMLValue(val)

                val = j.find(util.nspath_eval('gmx:CodeDefinition/gml32:identifier', namespaces))
                self.dictionaries[id]['entries'][id2]['identifier'] = util.testXMLValue(val)

                val = j.find(util.nspath_eval('gmx:CodeDefinition', namespaces)).attrib.get('codeSpace')
                self.dictionaries[id]['entries'][id2]['codespace'] = util.testXMLValue(val, True)

    def getcodelistdictionaries(self):
        return list(self.dictionaries.keys())

    def getcodedefinitionidentifiers(self, cdl):
        if cdl in self.dictionaries:
            ids = []
            for i in self.dictionaries[cdl]['entries']:
                ids.append(self.dictionaries[cdl]['entries'][i]['identifier'])
            return ids
        else:
            return None


class MD_FeatureCatalogueDescription(object):
    """Process gmd:MD_FeatureCatalogueDescription"""
    def __init__(self, fcd=None):
        if fcd is None:
            self.xml = None
            self.compliancecode = None
            self.language = []
            self.includedwithdataset = None
            self.featuretypenames = []
            self.featurecatalogues = []
        else:
            if hasattr(fcd, 'getroot'):  # standalone document
                self.xml = etree.tostring(fcd.getroot())
            else:  # part of a larger document
                self.xml = etree.tostring(fcd)

            self.compliancecode = None
            val = fcd.find(util.nspath_eval('gmd:complianceCode/gco:Boolean', namespaces))
            val = util.testXMLValue(val)
            if val is not None:
                self.compliancecode = util.getTypedValue('boolean', val)

            self.language = []
            for i in fcd.findall(util.nspath_eval('gmd:language/gco:CharacterString', namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.language.append(val)

            self.includedwithdataset = None
            val = fcd.find(util.nspath_eval('gmd:includedWithDataset/gco:Boolean', namespaces))
            val = util.testXMLValue(val)
            if val is not None:
                self.includedwithdataset = util.getTypedValue('boolean', val)

            self.featuretypenames = []
            for i in fcd.findall(util.nspath_eval('gmd:featureTypes/gco:LocalName', namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.featuretypenames.append(val)
            for i in fcd.findall(util.nspath_eval('gmd:featureTypes/gco:ScopedName', namespaces)):
                val = util.testXMLValue(i)
                if val is not None:
                    self.featuretypenames.append(val)

            self.featurecatalogues = []
            for i in fcd.findall(util.nspath_eval('gmd:featureCatalogueCitation', namespaces)):
                val = i.attrib.get('uuidref')
                val = util.testXMLValue(val, attrib=True)
                if val is not None:
                    self.featurecatalogues.append(val)


class FC_FeatureCatalogue(object):
    """Process gfc:FC_FeatureCatalogue"""
    def __init__(self, fc=None):
        if fc is None:
            self.xml = None
            self.identifier = None
            self.name = None
            self.versiondate = None
            self.producer = None
            self.featuretypes = []
        else:
            if hasattr(fc, 'getroot'):  # standalone document
                self.xml = etree.tostring(fc.getroot())
            else:  # part of a larger document
                self.xml = etree.tostring(fc)

            val = fc.attrib['uuid']
            self.identifier = util.testXMLValue(val, attrib=True)

            val = fc.find(util.nspath_eval('gmx:name/gco:CharacterString', namespaces))
            self.name = util.testXMLValue(val)

            val = fc.find(util.nspath_eval('gmx:versionDate/gco:Date', namespaces))
            self.versiondate = util.testXMLValue(val)

            if not self.versiondate:
                val = fc.find(util.nspath_eval('gmx:versionDate/gco:DateTime', namespaces))
                self.versiondate = util.testXMLValue(val)

            self.producer = None
            prod = fc.find(util.nspath_eval('gfc:producer/gmd:CI_ResponsibleParty', namespaces))
            if prod is not None:
                self.producer = CI_ResponsibleParty(prod)

            self.featuretypes = []
            for i in fc.findall(util.nspath_eval('gfc:featureType/gfc:FC_FeatureType', namespaces)):
                self.featuretypes.append(FC_FeatureType(i))


class FC_FeatureType(object):
    """Process gfc:FC_FeatureType"""
    def __init__(self, ft=None):
        if ft is None:
            self.xml = None
            self.identifier = None
            self.typename = None
            self.definition = None
            self.isabstract = None
            self.aliases = []
            self.attributes = []
        else:
            if hasattr(ft, 'getroot'):  # standalone document
                self.xml = etree.tostring(ft.getroot())
            else:  # part of a larger document
                self.xml = etree.tostring(ft)

            val = ft.attrib['uuid']
            self.identifier = util.testXMLValue(val, attrib=True)

            val = ft.find(util.nspath_eval('gfc:typeName/gco:LocalName', namespaces))
            self.typename = util.testXMLValue(val)

            val = ft.find(util.nspath_eval('gfc:definition/gco:CharacterString', namespaces))
            self.definition = util.testXMLValue(val)

            self.isabstract = None
            val = ft.find(util.nspath_eval('gfc:isAbstract/gco:Boolean', namespaces))
            val = util.testXMLValue(val)
            if val is not None:
                self.isabstract = util.getTypedValue('boolean', val)

            self.aliases = []
            for i in ft.findall(util.nspath_eval('gfc:aliases/gco:LocalName', namespaces)):
                self.aliases.append(util.testXMLValue(i))

            self.attributes = []
            for i in ft.findall(util.nspath_eval('gfc:carrierOfCharacteristics/gfc:FC_FeatureAttribute', namespaces)):
                self.attributes.append(FC_FeatureAttribute(i))


class FC_FeatureAttribute(object):
    """Process gfc:FC_FeatureAttribute"""
    def __init__(self, fa=None):
        if fa is None:
            self.xml = None
            self.membername = None
            self.definition = None
            self.code = None
            self.valuetype = None
            self.listedvalues = []
        else:
            if hasattr(fa, 'getroot'):  # standalone document
                self.xml = etree.tostring(fa.getroot())
            else:  # part of a larger document
                self.xml = etree.tostring(fa)

            val = fa.find(util.nspath_eval('gfc:memberName/gco:LocalName', namespaces))
            self.membername = util.testXMLValue(val)

            val = fa.find(util.nspath_eval('gfc:definition/gco:CharacterString', namespaces))
            self.definition = util.testXMLValue(val)

            val = fa.find(util.nspath_eval('gfc:code/gco:CharacterString', namespaces))
            self.code = util.testXMLValue(val)

            val = fa.find(util.nspath_eval('gfc:valueType/gco:TypeName/gco:aName/gco:CharacterString', namespaces))
            self.valuetype = util.testXMLValue(val)

            self.listedvalues = []
            for i in fa.findall(util.nspath_eval('gfc:listedValue/gfc:FC_ListedValue', namespaces)):
                self.listedvalues.append(FC_ListedValue(i))


class FC_ListedValue(object):
    """Process gfc:FC_ListedValue"""
    def __init__(self, lv=None):
        if lv is None:
            self.xml = None
            self.label = None
            self.code = None
            self.definition = None
        else:
            if hasattr(lv, 'getroot'):  # standalone document
                self.xml = etree.tostring(lv.getroot())
            else:  # part of a larger document
                self.xml = etree.tostring(lv)

            val = lv.find(util.nspath_eval('gfc:label/gco:CharacterString', namespaces))
            self.label = util.testXMLValue(val)

            val = lv.find(util.nspath_eval('gfc:code/gco:CharacterString', namespaces))
            self.code = util.testXMLValue(val)

            val = lv.find(util.nspath_eval('gfc:definition/gco:CharacterString', namespaces))
            self.definition = util.testXMLValue(val)


class MD_ImageDescription(object):
    """Process gmd:MD_ImageDescription"""
    def __init__(self, img_desc=None):
        self.type = 'image'
        self.bands = []

        if img_desc is None:
            self.attribute_description = None
            self.cloud_cover = None
            self.processing_level = None
            self.illumination_elevation_angle = None
            self.illumination_azimuth_angle = None
        else:
            val = img_desc.find(util.nspath_eval('gmd:attributeDescription/gco:RecordType', namespaces))
            self.attribute_description = util.testXMLValue(val)

            val = img_desc.find(util.nspath_eval('gmd:contentType/gmd:MD_CoverageContentTypeCode', namespaces))
            self.type = util.testXMLAttribute(val, 'codeListValue')

            val = img_desc.find(util.nspath_eval('gmd:cloudCoverPercentage/gco:Real', namespaces))
            self.cloud_cover = util.testXMLValue(val)

            val = img_desc.find(util.nspath_eval('gmd:illuminationElevationAngle/gco:Real', namespaces))
            self.illumination_elevation_angle = util.testXMLValue(val)

            val = img_desc.find(util.nspath_eval('gmd:illuminationAzimuthAngle/gco:Real', namespaces))
            self.illumination_azimuth_angle = util.testXMLValue(val)

            val = img_desc.find(util.nspath_eval(
                'gmd:processingLevelCode/gmd:RS_Identifier/gmd:code/gco:CharacterString', namespaces))
            self.processing_level = util.testXMLValue(val)

            for i in img_desc.findall(util.nspath_eval('gmd:dimension/gmd:MD_Band', namespaces)):
                bid = util.testXMLAttribute(i, 'id')
                self.bands.append(MD_Band(i, bid))


class MD_Band(object):
    """Process gmd:MD_Band"""
    def __init__(self, band, band_id=None):
        if band is None:
            self.id = None
            self.units = None
            self.min = None
            self.max = None
        else:
            self.id = band_id

            val = band.find(util.nspath_eval('gmd:units/gml:UnitDefinition/gml:identifier', namespaces))
            self.units = util.testXMLValue(val)

            val = band.find(util.nspath_eval('gmd:minValue/gco:Real', namespaces))
            self.min = util.testXMLValue(val)

            val = band.find(util.nspath_eval('gmd:maxValue/gco:Real', namespaces))
            self.max = util.testXMLValue(val)


class MI_AcquisitionInformation(object):
    """Process gmi:MI_AcquisitionInformation"""

    def __init__(self, acq=None):
        self.platforms = []

        for i in acq.findall(util.nspath_eval('gmi:platform/gmi:MI_Platform', namespaces)):
            self.platforms.append(MI_Platform(i))


class MI_Platform(object):
    """Process gmi:MI_Platform"""

    def __init__(self, plt=None):
        self.instruments = []

        if plt is None:
            self.identifier = None
            self.description = None
        else:
            val = plt.find(util.nspath_eval('gmi:identifier', namespaces))
            self.identifier = util.testXMLValue(val)

            val = plt.find(util.nspath_eval('gmi:description', namespaces))
            self.description = util.testXMLValue(val)

            for i in plt.findall(util.nspath_eval('gmi:instrument/gmi:MI_Instrument', namespaces)):
                self.instruments.append(MI_Instrument(i))


class MI_Instrument(object):
    """Process gmi:MI_Instrument"""

    def __init__(self, inst=None):
        if inst is None:
            self.identifier = None
            self.type = None
        else:
            val = inst.find(util.nspath_eval('gmi:identifier', namespaces))
            self.identifier = util.testXMLValue(val)

            val = inst.find(util.nspath_eval('gmi:type', namespaces))
            self.type = util.testXMLValue(val)
